# Builtin
import json
import unittest
import logging
import os
from collections import OrderedDict

# Internal
from nxt import DATA_STATE, nxt_node, nxt_layer
from nxt import stage
from nxt import nxt_path
from nxt.constants import GRAPH_VERSION
from nxt.nxt_node import INTERNAL_ATTRS
from nxt.session import Session
from nxt.nxt_layer import SAVE_KEY, META_DATA_KEY

path_logger = logging.getLogger(nxt_path.__name__)
path_logger.propagate = False


class StageGeneral(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        os.chdir(os.path.dirname(__file__))
        cls.stage = Session().load_file(filepath="./StageLoadTest.nxt")
        cls.comp_layer = cls.stage.build_stage()
        cls.runtime_layer = cls.stage.setup_runtime_layer(cls.comp_layer)

    def test_load_data_against_save_data(self):
        """Orders predefined save data and tests against save data generated
        by the stage.
        :return: None
        """
        self.expected_data = r"""{
    "version": "1.17",
    "alias": "StageLoadTest",
    "color": "#22728f",
    "mute": false,
    "solo": false,
    "meta_data": {
        "positions": {
            "/Types": [
                4,
                20
            ]
        },
        "<KeepMe>": {
            "foo": "bar"
        }
    },
    "nodes": {
        "/Types": {
            "enabled": true,
            "attrs": {
                "_bool": {
                    "type": "bool",
                    "value": "True"
                },
                "_dict": {
                    "type": "dict",
                    "value": "{}"
                },
                "_float": {
                    "type": "float",
                    "value": "0.5"
                },
                "_int": {
                    "type": "int",
                    "value": "123"
                },
                "_list": {
                    "type": "list",
                    "value": "[${_str1}, ${_str2}, '${_raw}_limb_pv', '${_raw}_limb_ik']"
                },
                "_none": {
                    "type": "NoneType"
                },
                "_raw": {
                    "type": "raw",
                    "value": "l"
                },
                "_str1": {
                    "type": "str",
                    "value": "'single'"
                },
                "_str2": {
                    "type": "str",
                    "value": "\"double\""
                },
                "_tuple": {
                    "type": "tuple",
                    "value": "(${_int}, ${_float})"
                }
            },
            "code": [
                "${_list}  # List",
                "${_tuple}  # Tuple",
                "${_str1}  # String with single quote",
                "${_str2}  # String with double quote ",
                "${_raw}  # Raw",
                "${_int}  # Int",
                "${_float}  # Float",
                "${_dict}  # Dict",
                "${_bool}  # Bool"
            ]
        }
    }
}"""
        print("Testing a save data dict against a "
              "literal dict of the loaded data")
        self.raw_save_data = self.stage.get_layer_save_data(0)
        self.save_data = json.dumps(self.raw_save_data, indent=4,
                                    separators=(',', ': '))
        self.assertEqual(self.expected_data, self.save_data)

    def test_lookup(self):
        print("testing lookup")
        print("Testing that `/Node` is a invalid namespace.")
        self.assertIsNone(self.comp_layer.lookup("/Node"))
        print("Testing that `/Types` is a valid namespace.")
        self.assertIsNotNone(self.comp_layer.lookup("/Types"))
        print("Testing that the node at `/Types` is named `Types`.")
        self.assertEqual('Types', getattr(self.comp_layer.lookup("/Types"),
                                          INTERNAL_ATTRS.NAME))

    def test_runtime_lookup(self):
        print("test runtime lookup")
        print("Testing that `/Node` is a invalid runtime namespace.")
        self.assertIsNone(self.runtime_layer.lookup("/Node"))
        print("Testing that `/Types` is a valid runtime namespace.")
        n = self.runtime_layer.lookup("/Types")
        self.assertIsNotNone(n)
        print("Testing that the runtime node at `/Types` is named `Types`.")
        self.assertEqual('Types', getattr(self.runtime_layer.lookup("/Types"),
                                          INTERNAL_ATTRS.NAME))


class StageParentPaths(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        print("SetUp StageInherit")
        os.chdir(os.path.dirname(__file__))
        cls.stage = Session().load_file(filepath="./StageInheritTest.nxt")
        cls.comp_layer = cls.stage.build_stage()
        cls.layer = cls.stage._sub_layers[0]
        cls.parent_path = '/parent_node'
        cls.parent_node = cls.comp_layer.lookup(cls.parent_path)
        cls.child_path = '/parent_node/child_node'
        cls.child_node = cls.comp_layer.lookup(cls.child_path)
        cls.node1 = cls.comp_layer.lookup('/node1')
        cls.node2 = cls.comp_layer.lookup('/node2')
        cls.node3 = cls.comp_layer.lookup('/node3')
        cls.node3_child = cls.comp_layer.lookup('/node3/child')
        cls.inst_tgt = cls.comp_layer.lookup('/inst_tgt_parent')
        cls.inst_tgt_child = cls.comp_layer.lookup('/inst_tgt_parent/child')

    def test_internal_attr_inherit(self):
        print("Test if internal attrs are inherited as expected.")
        print("Test if the child node has the attr "
              "`.{}`".format(INTERNAL_ATTRS.PARENT_PATH))
        self.assertIs(True, hasattr(self.child_node,
                                    INTERNAL_ATTRS.PARENT_PATH))
        print("Test if the child node `.{}` "
              "value is not None".format(INTERNAL_ATTRS.PARENT_PATH))
        self.assertIsNotNone(getattr(self.child_node,
                                     INTERNAL_ATTRS.PARENT_PATH))
        print("Test if the inherit node's name is really `parent_node`")
        parent_path = getattr(self.child_node, INTERNAL_ATTRS.PARENT_PATH)
        parent = self.layer.lookup(parent_path)
        name = getattr(parent, INTERNAL_ATTRS.NAME)
        self.assertEqual('parent_node', name)
        # COMPUTE INHERITANCE
        pn_expected_code = ['# Code']
        pn_actual_code = getattr(self.parent_node, INTERNAL_ATTRS.COMPUTE)
        self.assertEqual(pn_expected_code, pn_actual_code)
        cn_expected_code = []
        cn_actual_code = getattr(self.child_node, INTERNAL_ATTRS.COMPUTE)
        self.assertEqual(cn_expected_code, cn_actual_code)
        n1_expected_code = ['# BOTTOM']
        n1_actual_code = getattr(self.node1, INTERNAL_ATTRS.COMPUTE)
        self.assertEqual(n1_expected_code, n1_actual_code)
        n2_expected_code = ['']
        n2_actual_code = getattr(self.node2, INTERNAL_ATTRS.COMPUTE)
        self.assertEqual(n2_expected_code, n2_actual_code)
        inst_tgt_expected_code = ['# Parent code']
        inst_tgt_actual_code = getattr(self.inst_tgt, INTERNAL_ATTRS.COMPUTE)
        self.assertEqual(inst_tgt_expected_code, inst_tgt_actual_code)
        inst_tgt_expected_child_code = ['# Child code']
        inst_tgt_actual_child_code = getattr(self.inst_tgt_child,
                                                INTERNAL_ATTRS.COMPUTE)
        self.assertEqual(inst_tgt_expected_child_code,
                         inst_tgt_actual_child_code)
        # COMMENT INHERITANCE
        pn_expected_comment = 'Node - layer 1'
        pn_actual_comment = getattr(self.parent_node, INTERNAL_ATTRS.COMMENT)
        self.assertEqual(pn_expected_comment, pn_actual_comment)
        cn_expected_comment = None
        cn_actual_comment = getattr(self.child_node, INTERNAL_ATTRS.COMMENT)
        self.assertEqual(cn_expected_comment, cn_actual_comment)
        n1_expected_comment = 'Node1 - layer 0'
        n1_actual_comment = getattr(self.node1, INTERNAL_ATTRS.COMMENT)
        self.assertEqual(n1_expected_comment, n1_actual_comment)
        n2_expected_comment = 'Node2 - layer 0'
        n2_actual_comment = getattr(self.node2, INTERNAL_ATTRS.COMMENT)
        self.assertEqual(n2_expected_comment, n2_actual_comment)
        # START POINT INHERITANCE
        # TODO: When save files are flat test start points
        # INSTANCE INHERITANCE
        n2_expected_inst_path = '/Instance/Path'
        n2_actual_inst_path = getattr(self.node2, INTERNAL_ATTRS.INSTANCE_PATH)
        self.assertEqual(n2_expected_inst_path, n2_actual_inst_path)
        # EXEC INHERITANCE
        n2_expected_exec_in = '/node1'
        n2_actual_exec_in = getattr(self.node2, INTERNAL_ATTRS.EXECUTE_IN)
        self.assertEqual(n2_expected_exec_in, n2_actual_exec_in)
        # ENABLED INHERITANCE
        n2_expected_enabled = True
        n2_actual_enabled = getattr(self.node2, INTERNAL_ATTRS.ENABLED)
        self.assertEqual(n2_expected_enabled, n2_actual_enabled)
        parent_expected_enabled = False
        parent_actual_enabled = getattr(self.parent_node,
                                        INTERNAL_ATTRS.ENABLED)
        self.assertEqual(parent_expected_enabled, parent_actual_enabled)
        child_expected_enabled = True
        child_actual_enabled = getattr(self.child_node, INTERNAL_ATTRS.ENABLED)
        self.assertEqual(child_expected_enabled, child_actual_enabled)
        n3_expected_enabled = False
        n3_actual_enabled = getattr(self.node3, INTERNAL_ATTRS.ENABLED)
        self.assertEqual(n3_expected_enabled, n3_actual_enabled)
        n3_child_expected_enabled = None
        n3_child_actual_enabled = getattr(self.node3_child, INTERNAL_ATTRS.ENABLED)
        self.assertEqual(n3_child_expected_enabled, n3_child_actual_enabled)

    def test_inherit_attrs(self):
        print("Testing if child node inherits attrs from its parent...")
        print("Checking that the `.parent` attr from the parent node is NOT "
              "local to the child node.")
        child_locals = nxt_layer.get_node_local_attr_names(self.child_path,
                                                           [self.layer])
        self.assertNotIn('parent', child_locals)
        print("Checking that the `.parent` attr from the parent node IS "
              "local to the parent node.")
        parent_locals = nxt_layer.get_node_local_attr_names(self.parent_path,
                                                            [self.layer])
        self.assertIn('parent', parent_locals)
        # Test that code is inherited across layers
        print("Test that the code IS inherited from the LOWER LAYER.")
        self.assertEqual(["# BOTTOM"], getattr(self.node1,
                                               INTERNAL_ATTRS.COMPUTE))
        self.assertNotEqual(["# BOTTOM2"], getattr(self.node2,
                                                   INTERNAL_ATTRS.COMPUTE))
        # Test that session layer is created
        print("Generating session layer and runtime nodes...")
        self.runtime_layer = self.stage.execute('/parent_node')
        print("Confirm `runtime_layer` is not None.")
        self.assertIsNotNone(self.runtime_layer)
        # Test that node spec has the same name as the runtime node
        runtime_parent = self.runtime_layer.lookup('/parent_node')
        runtime_parent_children = self.runtime_layer.children('/parent_node',
                                                              ordered=True)
        rt_node1 = self.runtime_layer.lookup('/node1')
        rt_node2 = self.runtime_layer.lookup('/node2')
        print("Checking that runtime PARENT node name matches editor PARENT "
              "node name.")
        name = getattr(self.parent_node, INTERNAL_ATTRS.NAME)
        rt_name = getattr(runtime_parent, INTERNAL_ATTRS.NAME)
        self.assertEqual(name, rt_name)
        runtime_child = runtime_parent_children[0]
        print("Checking that runtime CHILD node name matches editor CHILD "
              "node name.")
        name = getattr(self.child_node, INTERNAL_ATTRS.NAME)
        rt_name = getattr(runtime_child, INTERNAL_ATTRS.NAME)
        self.assertEqual(name, rt_name)
        print("Test that the local attr from the parent node IS accessible"
              "from the RUNTIME child node via `self`.")
        self.assertIs(True, hasattr(runtime_child, 'parent'))
        print("Test that the local attr from the parent node is IS accessible"
              "from the EDITOR child node via `self`.")
        self.assertEquals(True, hasattr(self.child_node, 'parent'))
        # Test that code is not inherited
        print("Test that child code is NOT inherited FROM the PARENT.")
        self.assertEqual(["# Code"], getattr(runtime_parent,
                                             INTERNAL_ATTRS.COMPUTE))
        self.assertNotEqual(["# Code"], getattr(runtime_child,
                                                INTERNAL_ATTRS.COMPUTE))
        # Test that code is inherited across layers
        print("Test that the code IS inherited from the LOWER LAYER.")
        self.assertEqual(["# BOTTOM"], getattr(rt_node1,
                                               INTERNAL_ATTRS.COMPUTE))
        self.assertNotEqual(["# BOTTOM2"], getattr(rt_node2,
                                                   INTERNAL_ATTRS.COMPUTE))
        print("Test that attrs are overloaded top down.")
        expected = "top"
        actual = getattr(self.node2, 'attr')
        self.assertEqual(expected, actual)

    def test_parent_function(self):
        print("Test that paths are not broken when nodes with similar names "
              "are parented together.")
        target_layer = self.stage.top_layer
        node = target_layer.lookup('/parent_node')
        self.stage.parent_nodes(nodes=[node], parent_path='/node2',
                                layer=target_layer)
        self.comp_layer = self.stage.build_stage()
        expected_path = '/node2/parent_node/child_node'
        display_node = self.comp_layer.lookup('/node2/parent_node/child_node')
        self.assertIsNotNone(display_node)
        display_name = getattr(display_node, INTERNAL_ATTRS.NAME)
        display_parent_path = getattr(display_node, INTERNAL_ATTRS.PARENT_PATH)
        actual_path = nxt_path.join_node_paths(display_parent_path,
                                               display_name)
        self.assertEqual(expected_path, actual_path)


class StageInstance1(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        print("Class SetUp StageInstance")
        os.chdir(os.path.dirname(__file__))
        cls.stage = Session().load_file(filepath="./StageInstanceTest.nxt")
        cls.comp_layer = cls.stage.build_stage()
        cls.layer = cls.stage._sub_layers[0]
        # Lookup the nodes we are instancing from/to (source, target)
        #       Test1
        cls.inst_source1_path = '/inst_source1'
        cls.inst_source1 = cls.comp_layer.lookup(cls.inst_source1_path)
        cls.inst_target1_path = '/inst_target1'
        cls.inst_target1 = cls.comp_layer.lookup(cls.inst_target1_path)
        cls.inst_source1_child_path = '/inst_source1/inst_source1_child'
        cls.inst_source1_child = cls.comp_layer.lookup(cls.inst_source1_child_path)
        cls.inst_target1_child_path = '/inst_target1/inst_source1_child'
        cls.inst_target1_child = cls.comp_layer.lookup(cls.inst_target1_child_path)
        #       Test 2
        cls.inst_source2_path = '/inst_source2'
        cls.inst_source2 = cls.comp_layer.lookup(cls.inst_source2_path)
        cls.inst_source2_child_path = '/inst_source2/inst_source2_child'
        cls.inst_source2_child = cls.comp_layer.lookup(cls.inst_source2_child_path)
        cls.inst_source2_middle_path = '/inst_source2_middle'
        cls.inst_source2_middle = cls.comp_layer.lookup(cls.inst_source2_middle_path)
        cls.inst_source2_middle_child_path = '/inst_source2_middle/inst_source2_middle_child'
        cls.inst_source2_middle_child = cls.comp_layer.lookup(cls.inst_source2_middle_child_path)

        cls.inst_target2_path = '/inst_target2'
        cls.inst_target2 = cls.comp_layer.lookup(cls.inst_target2_path)
        cls.inst_target2_child_path = '/inst_source2_middle/inst_source2_child'
        cls.inst_target2_child = cls.comp_layer.lookup(cls.inst_target2_child_path)
        cls.inst_target2_middle_child_path = '/inst_source2_middle/inst_source2_middle_child'
        cls.inst_target2_middle_child = cls.comp_layer.lookup(cls.inst_target2_middle_child_path)

        # Edgecase test
        cls.top_node_path = '/top'
        cls.top_node = cls.comp_layer.lookup(cls.top_node_path)
        cls.inst_from_node_path = '/instFrom'
        cls.inst_from_node = cls.comp_layer.lookup(cls.inst_from_node_path)
        cls.inst_to_node_path = '/top/instTo'
        cls.inst_to_node = cls.comp_layer.lookup(cls.inst_to_node_path)

        # Setup the session layer which generates runtime nodes
        cls.runtime_layer = cls.stage.setup_runtime_layer(cls.comp_layer)
        # Lookup all the runtime nodes
        #       Test 1
        cls.rt_inst_source1 = cls.runtime_layer.lookup(cls.inst_source1_path)
        cls.rt_inst_target1 = cls.runtime_layer.lookup(cls.inst_target1_path)
        cls.rt_inst_source1_child = cls.runtime_layer.lookup(cls.inst_source1_child_path)
        cls.rt_inst_target1_child = cls.runtime_layer.lookup(cls.inst_target1_child_path)
        cls.rt_inst_target1_real_child_path = '/inst_target1/RealChild'
        cls.rt_inst_target1_real_child = cls.runtime_layer.lookup(cls.rt_inst_target1_real_child_path)
        #       Test 2
        cls.rt_inst_source2 = cls.runtime_layer.lookup(cls.inst_source2_path)
        cls.rt_inst_source2_child = cls.runtime_layer.lookup(cls.inst_source2_child_path)

        cls.rt_inst_source2_middle = cls.runtime_layer.lookup(cls.inst_source2_middle_path)
        cls.rt_inst_source2_middle_child = cls.runtime_layer.lookup(cls.inst_source2_middle_child_path)
        cls.rt_inst_source2_middle_inst_child_path = '/inst_source2_middle/inst_source2_child'
        cls.rt_inst_source2_middle_inst_child = cls.runtime_layer.lookup(cls.rt_inst_source2_middle_inst_child_path)

        cls.rt_inst_target2 = cls.runtime_layer.lookup(cls.inst_target2_path)
        cls.rt_inst_target2_middle_child = cls.runtime_layer.lookup(cls.inst_target2_middle_child_path)
        cls.rt_inst_target2_child = cls.runtime_layer.lookup(cls.inst_target2_child_path)

    def test_editor_time1(self):
        print("Testing instancing scenario 1 at EDITOR time...")
        print("Test the editor time instance source is set as the "
              "`._instance` for the target.")
        inst, _ = self.stage.safe_get_node_instance(self.inst_target1,
                                                    self.comp_layer)
        self.assertIs(self.inst_source1, inst)
        # And do the same for the child node.
        print("Same test for child node.")
        tgt_children = self.comp_layer.children(self.inst_target1_path,
                                                ordered=True)
        child_inst, _ = self.stage.safe_get_node_instance(tgt_children[1],
                                                          self.comp_layer)
        self.assertIs(self.inst_source1_child, child_inst)
        # The child node should be a proxy node
        print("Test that child node is a "
              "`.{}` node.".format(INTERNAL_ATTRS.PROXY))
        self.assertIs(True, getattr(tgt_children[1], INTERNAL_ATTRS.PROXY))
        # Test if nodes have all the expected children
        print("Test that the instance SOURCE have the expected number of "
              "children (1).")
        src_children = self.comp_layer.children(self.inst_source1_path)
        self.assertEqual(1, len(src_children))
        print("Test that the instance TARGET have the expected number of "
              "children (2).")
        self.assertEqual(2, len(tgt_children))

    def test_editor_time2(self):
        print("Testing instancing scenario 2 at EDITOR time...")
        # Test case 2 editor time has all the expected children
        print("Testing number of children for the three trees in this "
              "scenario.")
        children = self.comp_layer.children(self.inst_source2_path)
        self.assertEqual(1, len(children))
        children = self.comp_layer.children(self.inst_source2_middle_path)
        self.assertEqual(2, len(children))
        children = self.comp_layer.children(self.inst_target2_path)
        self.assertEqual(2, len(children))
        print("Testing that all resolved codes match expected literals.")
        c = self.stage.get_node_code_lines(self.inst_source2,
                                              self.comp_layer,
                                              data_state=DATA_STATE.RESOLVED)
        self.assertEqual(["# Instance Root Code"], c)
        c = self.stage.get_node_code_lines(self.inst_source2_child,
                                              self.comp_layer,
                                              data_state=DATA_STATE.RESOLVED)
        self.assertEqual(["# Instance Code"], c)
        c = self.stage.get_node_code_lines(self.inst_source2_middle,
                                              self.comp_layer,
                                              data_state=DATA_STATE.RESOLVED)
        self.assertEqual(["# Instance Root Code"], c)
        c = self.stage.get_node_code_lines(self.inst_source2_middle_child,
                                              self.comp_layer,
                                              data_state=DATA_STATE.RESOLVED)
        self.assertEqual(["# Instance Middle Code"], c)
        c = self.stage.get_node_code_lines(self.inst_target2,
                                              self.comp_layer,
                                              data_state=DATA_STATE.RESOLVED)
        self.assertEqual(["# Instance Root Code"], c)
        c = self.stage.get_node_code_lines(self.inst_target2_middle_child,
                                              self.comp_layer,
                                              data_state=DATA_STATE.RESOLVED)
        self.assertEqual(["# Instance Middle Code"], c)
        c = self.stage.get_node_code_lines(self.inst_target2_child,
                                              self.comp_layer,
                                              data_state=DATA_STATE.RESOLVED)
        self.assertEqual(["# Instance Code"], c)

    def test_runtime_setup1(self):
        func = self.stage.get_node_code_lines
        print("Testing instancing scenario 1 at RUNTIME...")
        print("Testing that expected attrs are on the runtime target node.")
        self.assertIs(True, hasattr(self.rt_inst_target1, "InstSourceAttr"))
        print("Testing the value is the expected one.")
        self.assertEqual('Hello', getattr(self.rt_inst_target1,
                                          "InstSourceAttr"))
        print("Testing the instance target code resolves as expected, "
              "uses a local and remote attr ref.")
        rt_real_child_resolved_code = func(self.rt_inst_target1_real_child,
                                              layer=self.runtime_layer,
                                              data_state=DATA_STATE.RESOLVED)
        self.assertEqual(["\"Hello World\""], rt_real_child_resolved_code)
        # Test if inst source code is the same as target code
        print("Test if inst source code IS the same as target code.")
        rt_inst_target_code = func(self.rt_inst_target1,
                                      layer=self.runtime_layer,
                                      data_state=DATA_STATE.RESOLVED)
        expected_code = getattr(self.rt_inst_source1, INTERNAL_ATTRS.COMPUTE)
        self.assertEqual(expected_code, rt_inst_target_code)
        print("Test if inst SOURCE CHILD code is the same as the TARGET "
              "inst CHILD code")
        rt_inst_target_child_code = func(self.rt_inst_target1_child,
                                            layer=self.runtime_layer,
                                            data_state=DATA_STATE.RESOLVED)
        expected_code = getattr(self.rt_inst_source1_child,
                                   INTERNAL_ATTRS.COMPUTE)
        self.assertEqual(expected_code, rt_inst_target_child_code)

    def test_runtime_setup2(self):
        func = self.stage.get_node_code_lines
        print("Testing instancing scenario 2 at RUNTIME...")
        # Test case 2 runtime has all the expected children
        print("Testing number of children for the three runtime trees in "
              "this scenario.")
        children = self.runtime_layer.children(self.inst_source2_path)
        self.assertEqual(1, len(children))
        children = self.runtime_layer.children(self.inst_source2_middle_path)
        self.assertEqual(2, len(children))
        children = self.runtime_layer.children(self.inst_target2_path)
        self.assertEqual(2, len(children))
        print("Testing that all runtime resolved codes match expected "
              "literals.")
        c = func(self.rt_inst_source2, layer=self.runtime_layer,
                 data_state=DATA_STATE.RESOLVED)
        self.assertEqual(["# Instance Root Code"], c)
        c = func(self.inst_source2_child, layer=self.runtime_layer,
                 data_state=DATA_STATE.RESOLVED)
        self.assertEqual(["# Instance Code"], c)
        c = func(self.rt_inst_source2_middle, layer=self.runtime_layer,
                 data_state=DATA_STATE.RESOLVED)
        self.assertEqual(["# Instance Root Code"], c)
        c = func(self.rt_inst_source2_middle_child, layer=self.runtime_layer,
                 data_state=DATA_STATE.RESOLVED)
        self.assertEqual(["# Instance Middle Code"], c)
        c = func(self.rt_inst_target2, layer=self.runtime_layer,
                 data_state=DATA_STATE.RESOLVED)
        self.assertEqual(["# Instance Root Code"], c)
        c = func(self.rt_inst_target2_middle_child, layer=self.runtime_layer,
                 data_state=DATA_STATE.RESOLVED)
        self.assertEqual(["# Instance Middle Code"], c)
        c = func(self.rt_inst_target2_child, layer=self.runtime_layer,
                 data_state=DATA_STATE.RESOLVED)
        self.assertEqual(["# Instance Code"], c)
        print("Test expected attrs are accessible via self.")
        # Test attrs are accessible via self
        self.assertIs(True, hasattr(self.rt_inst_source2_child, "attr0"))
        self.assertIs(True, hasattr(self.rt_inst_source2_child, "attr0_1"))
        # Middle real child
        self.assertIs(True, hasattr(self.rt_inst_source2_middle_child, "attr0"))
        self.assertIs(True, hasattr(self.rt_inst_source2_middle_child, "attr1"))
        self.assertIs(True, hasattr(self.rt_inst_source2_middle_child,
                                    "attr1_1"))
        # Middle inst child
        self.assertIs(True, hasattr(self.rt_inst_source2_middle_inst_child,
                                    "attr0"))
        self.assertIs(True, hasattr(self.rt_inst_source2_middle_inst_child,
                                    "attr1"))
        self.assertIs(True, hasattr(self.rt_inst_source2_middle_inst_child,
                                    "attr0_1"))
        # Target
        self.assertIs(True, hasattr(self.rt_inst_target2, "attr0"))
        self.assertIs(True, hasattr(self.rt_inst_target2, "attr1"))
        # Target inst source child (deepest instance child)
        self.assertIs(True, hasattr(self.rt_inst_target2_child, "attr0"))
        self.assertIs(True, hasattr(self.rt_inst_target2_child, "attr1"))
        self.assertIs(True, hasattr(self.rt_inst_target2_child, "attr0_1"))
        # Target middle inst child
        self.assertIs(True, hasattr(self.rt_inst_target2_middle_child,
                                    "attr1_1"))
        self.assertIs(True, hasattr(self.rt_inst_target2_middle_child,
                                    "attr1"))
        self.assertIs(True, hasattr(self.rt_inst_target2_middle_child,
                                    "attr0"))

    def test_name_edege_case1(self):
        print("Test that instance nodes parent to their parent and not their "
              "grandparent node.")
        top_children = self.comp_layer.children(self.top_node_path)
        top_child_names = [getattr(n, INTERNAL_ATTRS.NAME)
                           for n in top_children]
        self.assertEqual(['instTo'], top_child_names)
        source_children = self.comp_layer.children(self.inst_from_node_path,
                                                   ordered=True)
        source_child_names = [getattr(n, INTERNAL_ATTRS.NAME)
                              for n in source_children]
        target_children = self.comp_layer.children(self.inst_to_node_path,
                                                   ordered=True)
        target_child_names = [getattr(n, INTERNAL_ATTRS.NAME)
                              for n in target_children]
        self.assertEqual(source_child_names, target_child_names)

    def test_ns_depth_case1(self):
        print('Test that shallow namespaces correctly merge with deep.')
        node = self.comp_layer.lookup('/rig_workflow/file_io/importers/'
                                      'joint_importers/joints_layout/skeleton/'
                                      'connect_twist/joints')
        self.assertIsNone(node)
        node = self.comp_layer.lookup('/rig_workflow/file_io/importers/'
                                      'joint_importers/joints_layout/skeleton/'
                                      'connect_twist/joints/connect_twist')
        self.assertIsNone(node)
        node = self.comp_layer.lookup('/rig_workflow/file_io/importers/'
                                      'joint_importers/joints_layout/skeleton/'
                                      'connect_twist')
        self.assertIsNotNone(node)
        node = self.comp_layer.lookup('/rig_workflow/file_io/importers/'
                                      'joint_importers/joints_layout/skeleton/'
                                      'connect_twist/left/node')
        self.assertIsNotNone(node)


    def test_instance_child_attr_overload(self):
        '''Test that an instance source child can overload its parent attr
        value.'''
        print("Testing that an instance source child can overload its parent "
              "attr value.")
        proxy_node4 = self.comp_layer.lookup('/inst_target4/inst_source4_child')
        expected_code4 = "\"The base should be 1 is it? base == 1\""
        state = DATA_STATE.RESOLVED
        actual_code = self.stage.get_node_code_string(proxy_node4,
                                                            self.comp_layer,
                                                            state)
        self.assertEqual(expected_code4, actual_code)
        expected_attr = '1'
        actual_attr = self.stage.get_node_attr_value(proxy_node4, 'base',
                                                     self.comp_layer)
        self.assertEqual(expected_attr, actual_attr)
        print("Testing that the SOURCE CHILD overload is cached correctly.")
        rt_layer = self.stage.execute('/inst_source4')
        path = '/inst_target4/inst_source4_child'
        rt_proxy_node4 = rt_layer.cache_layer.lookup(path)
        rt_actual4 = getattr(rt_proxy_node4, INTERNAL_ATTRS.CACHED_CODE)
        self.assertEqual(expected_code4, rt_actual4)

    def test_instance_parent_attr_overload(self):
        print("Testing that an instance target can overload its "
              "child's attr value (1 level deep).")
        proxy5_path = '/inst_target5/inst_source5_child'
        proxy_node5 = self.comp_layer.lookup(proxy5_path)
        expected_attr = '5'
        actual_attr = self.stage.get_node_attr_value(proxy_node5, 'base',
                                                     self.comp_layer)
        self.assertEqual(expected_attr, actual_attr)
        print("Testing that an instance target can overload its "
              "child's attr value (2 levels deep).")
        proxy6_path = '/inst_target6/RealChild/inst_source6_child'
        proxy_node6 = self.comp_layer.lookup(proxy6_path)
        expected_attr = 'B'
        actual_attr = self.stage.get_node_attr_value(proxy_node6, 'base',
                                                     self.comp_layer)
        self.assertEqual(expected_attr, actual_attr)
        print("Testing that the TARGET PARENT overload is cached correctly "
              "(1 level deep).")
        rt_layer = self.stage.execute('/inst_source5')
        rt_proxy_node5 = rt_layer.cache_layer.lookup(proxy5_path)
        rt_actual5 = getattr(rt_proxy_node5, INTERNAL_ATTRS.CACHED_CODE)
        expected_code5 = "\"The base should be 5 is it? base == 5\""
        self.assertEqual(expected_code5, rt_actual5)
        print("Testing that the TARGET PARENT overload is cached correctly "
              "(2 levels deep).")
        rt_layer = self.stage.execute('/inst_source6')
        rt_proxy_node6 = rt_layer.cache_layer.lookup(proxy6_path)
        rt_actual6 = getattr(rt_proxy_node6, INTERNAL_ATTRS.CACHED_CODE)
        expected_code5 = "\"The base should be B is it? base == B\""
        self.assertEqual(expected_code5, rt_actual6)

    def test_instance_mid_to_top(self):
        print("Test that nodes can instance from a non-root to a root.")
        expected_children = ['c1', 'c2']
        rt = self.comp_layer.RETURNS.Path
        actual_children = self.comp_layer.children('/mid_tgt',
                                                   return_type=rt, ordered=True)
        actual_children = [nxt_path.node_name_from_node_path(p) for p in
                           actual_children]
        self.assertEqual(expected_children, actual_children)
        print("Test that instance path is correct.")
        expected_inst_path = '/top2/mid'
        node = self.comp_layer.lookup('/mid_tgt')
        ip = INTERNAL_ATTRS.INSTANCE_PATH
        actual_inst_path = self.stage.get_node_attr_value(node, ip,
                                                          self.comp_layer)
        self.assertEqual(expected_inst_path, actual_inst_path)


class StageInstance2(unittest.TestCase):

    @classmethod
    def setUp(cls):
        print("Class SetUp StageInstance")
        os.chdir(os.path.dirname(__file__))
        cls.stage = Session().load_file(filepath="./StageInstanceAcrossLayersTest.nxt")
        cls.comp_layer = cls.stage.build_stage()
        cls.layer = cls.stage._sub_layers[0]
        # Lookup the nodes we are instancing from/to (source, target)
        #       Test1
        cls.inst_source3_path = '/inst_source3'
        cls.inst_source3 = cls.comp_layer.lookup(cls.inst_source3_path)
        cls.inst_target3_path = '/inst_target3'
        cls.inst_target3 = cls.comp_layer.lookup(cls.inst_target3_path)
        cls.inst_target3_child_path = '/inst_target3/RealChild'
        cls.inst_target3_child = cls.comp_layer.lookup(cls.inst_target3_child_path)
        # Setup the session layer which generates runtime nodes
        cls.runtime_layer = cls.stage.setup_runtime_layer(cls.comp_layer)
        # Lookup all the runtime nodes
        #       Test 1
        cls.rt_inst_source3 = cls.runtime_layer.lookup('/inst_source3')
        cls.rt_inst_target3 = cls.runtime_layer.lookup('/inst_target3')
        cls.rt_inst_target3_child = cls.runtime_layer.lookup('/inst_target3/RealChild')

    def test_editor_time1(self):
        print("Testing instancing case 2 scenario 1 at EDITOR time...")
        print("Test that the expected number of children exist on the "
              "instance source.")
        children = self.comp_layer.children(self.inst_source3_path)
        self.assertEqual(2, len(children))
        print("Test that the expected number of children exist on the "
              "instance target")
        children = self.comp_layer.children(self.inst_target3_path)
        self.assertEqual(2, len(children))

    def test_runtime_time1(self):
        print("Testing instancing case 2 scenario 1 at RUNTIME time...")
        print("Test that the expected number of children exist on the "
              "instance source.")
        children = self.runtime_layer.children(self.inst_source3_path)
        self.assertEqual(2, len(children))
        print("Test that the expected number of children exist on the "
              "instance target")
        children = self.runtime_layer.children(self.inst_source3_path)
        self.assertEqual(2, len(children))

    def test_editor_add_node_hierarchy(self):
        print("Check that node is proxy")
        parent_path = '/inst_target3'
        new_node_path = '/inst_target3/RealChild'
        new_node = self.comp_layer.lookup(new_node_path)
        self.assertIs(True, getattr(new_node, INTERNAL_ATTRS.PROXY))
        print("Create node hierarchy")
        node_hierarchy = ['inst_target3', 'RealChild']
        top_layer = self.stage.top_layer
        comp_layer = self.comp_layer
        node_table, dirty = self.stage.add_node_hierarchy(node_hierarchy,
                                                          parent=None,
                                                          layer=top_layer)
        self.comp_layer = self.stage.build_stage()
        print("Check that node is NOT proxy")
        new_node = self.comp_layer.lookup('/inst_target3/RealChild')
        self.assertIs(False, getattr(new_node, INTERNAL_ATTRS.PROXY))
        print("Check dirty map")
        self.assertEqual([parent_path, new_node_path], dirty)

    def test_localizing_proxy_child(self):
        print("Check that proxy node dose not get its parent's compute when "
              "localized.")
        parent_path = '/inst_source3'
        node_name = 'inst_source1_child'
        node_path = nxt_path.join_node_paths(parent_path, node_name)
        parent_node = self.stage.top_layer.lookup('/inst_source3')
        self.stage.add_node(node_name, data={}, parent=parent_node,
                            layer=self.stage.top_layer,
                            comp_layer=self.comp_layer, fix_names=False)
        node = self.comp_layer.lookup(node_path)
        inst_src_path = '/inst_target1/inst_source1_child'
        inst_src = self.comp_layer.lookup(inst_src_path)
        expected = getattr(inst_src, INTERNAL_ATTRS.COMPUTE)
        actual = getattr(node, INTERNAL_ATTRS.COMPUTE)
        self.assertEqual(expected, actual)


class StageInstance3(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.stage = Session().load_file(filepath="./StageInstanceTest_Layer0.nxt")
        cls.comp_layer = cls.stage.build_stage()

    def test_localized_by_code(self):
        """Tests the when a node is localized by setting its code the comp
        node is properly updated."""
        print('Testing setting code on proxy node properly comps.')
        new_code = ['123']
        self.stage.add_node(name='fk', data={'code': new_code},
                            parent='/Character/build/legs/left/create',
                            layer=0, comp_layer=self.comp_layer,
                            fix_names=False)
        full_path = '/Character/build/legs/left/create/fk'
        new_node = self.comp_layer.lookup(full_path)
        live_code = getattr(new_node, INTERNAL_ATTRS.COMPUTE)
        # Test the code is set
        self.assertIs(new_code, live_code)
        layer = self.stage.top_layer
        spec_node = layer.lookup(full_path)
        self.stage.delete_node(spec_node, layer=layer,
                               comp_layer=self.comp_layer,
                               remove_layer_data=False)
        comp_node = self.comp_layer.lookup(full_path)
        live_code = getattr(new_node, INTERNAL_ATTRS.COMPUTE)
        # Test the code is gone
        self.assertEqual([], live_code)


    def test_deep_localized_node(self):
        """Tests that if a node, deep in a hierarchy, is localized and then
        instanced, does not result in the instances getting mangled into
        implied nodes.
        """
        print('Testing deep localized instance node does not end up implied.')
        real_node = self.comp_layer.lookup('/left/leg2/joints/upper')
        self.assertIsNotNone(real_node)

    def test_incidental_node_creation(self):
        """Test creating a node via adding an attr on a layer lower than
        the one who calls for the instance.
        """
        print("Test that incidentally creating a node lower than the display "
              "layer doesn't crash")
        attr_data = {'attrs': {'attr1': {'value': 'asdf'}}, 'name': 'upper'}
        node, _ = self.stage.add_node(name='upper', data=attr_data,
                                      parent='/leg/create/fk/controls',
                                      layer=2, comp_layer=self.comp_layer,
                                      fix_names=False)
        self.assertIsNotNone(node)

    def test_name_clash(self):
        """Test that child nodes with the same name as their parent "
              "instance correctly."""
        print("Test that there are the expected number of descendants (2)")
        children = self.comp_layer.children('/node')
        self.assertEqual(2, len(children))
        children = self.comp_layer.children('/node1')
        self.assertEqual(2, len(children))
        node_descendants = self.comp_layer.descendants('/node')
        node1_descendants = self.comp_layer.descendants('/node1')
        self.assertEqual(2, len(node_descendants))
        self.assertEqual(2, len(node1_descendants))

    def test_2_deep_instances(self):
        """Test that instances of instances have the expected children."""
        d = self.comp_layer.descendants('/leg')
        print("Test that there are the expected number of descendants (7)")
        self.assertEqual(7, len(d))
        upper_node = self.comp_layer.lookup('/leg/create/fk/controls/upper')
        lower_node = self.comp_layer.lookup('/leg/create/fk/controls/lower')
        print("Testing the node `/leg/create/fk/controls/upper` exists.")
        self.assertIsNotNone(upper_node)
        print("Testing the node `/leg/create/fk/controls/lower` exists.")
        self.assertIsNotNone(lower_node)
        upper_create_node = self.comp_layer.lookup('/leg/create/fk/controls/upper/create')
        lower_create_node = self.comp_layer.lookup('/leg/create/fk/controls/lower/create')
        print("Testing the node `/leg/create/fk/controls/upper/create` "
              "exists.")
        self.assertIsNotNone(upper_create_node)
        print("Testing the node `/leg/create/fk/controls/lower/create` "
              "exists.")
        self.assertIsNotNone(lower_create_node)

    def test_3_deep_instances(self):
        """Test instances of instances as well as twin instancing."""
        d = self.comp_layer.descendants('/Character')
        print("Test that there are the expected number of descendants (18)")
        self.assertEqual(18, len(d))
        left_node = self.comp_layer.lookup('/Character/build/legs/left')
        right_node = self.comp_layer.lookup('/Character/build/legs/right')
        print("Testing the node `/Character/build/legs/left` exists.")
        self.assertIsNotNone(left_node)
        print("Testing the node `/Character/build/legs/right` (twin of "
              "`..left`) exists.")
        self.assertIsNotNone(right_node)
        left_control = self.comp_layer.lookup('/Character/build/legs/left/create/fk/controls/upper/create')
        right_control = self.comp_layer.lookup('/Character/build/legs/right/create/fk/controls/upper/create')
        print("Testing the node "
              "`/Character/build/legs/left/create/fk/controls/upper/create` "
              "(3 level deep instance) exists.")
        self.assertIsNotNone(left_control)
        print("Testing the node "
              "`/Character/build/legs/left/create/fk/controls/upper/create` "
              "(3 level deep instance of twin) exists.")
        self.assertIsNotNone(right_control)

    def test_4_deep_instances(self):
        """Test instancing an entire hierarchy that contains 3 level deep
        instances as well as twins."""
        d = self.comp_layer.descendants('/another')
        print("Test that there are the expected number of descendants (18)")
        self.assertEqual(18, len(d))
        left_node = self.comp_layer.lookup('/another/build/legs/left')
        right_node = self.comp_layer.lookup('/another/build/legs/right')
        print("Testing the node `/another./uild/legs/left` (4 level deep "
              "instance) exists.")
        self.assertIsNotNone(left_node)
        print("Testing the node `/another/build/legs/right` (4 level deep "
              "instance) exists.")
        self.assertIsNotNone(right_node)
        left_control = self.comp_layer.lookup('/another/build/legs/left/create/fk/controls/upper/create')
        right_control = self.comp_layer.lookup('/another/build/legs/right/create/fk/controls/upper/create')
        print("Testing the node "
              "`/another/build/legs/right/create/fk/controls/upper/create` ("
              "4 level deep instance) exists.")
        self.assertIsNotNone(left_control)
        print("Testing the node "
              "`/another/build/legs/right/create/fk/controls/upper/create` ("
              "4 level deep instance) exists.")
        self.assertIsNotNone(right_control)

    def test_instance_node_on_layer_above(self):
        """Test that the proper instance nodes are created for a node who's
        instance path can not be resolved on its or any layer below itself."""
        d = self.comp_layer.descendants('/dummy')
        print("Test that there are the expected number of descendants (18)")
        self.assertEqual(18, len(d))
        left_node = self.comp_layer.lookup('/dummy/build/legs/left')
        right_node = self.comp_layer.lookup('/dummy/build/legs/right')
        print("Testing the node `/dummy/build/legs/left` exists.")
        self.assertIsNotNone(left_node)
        print("Testing the node `/dummy/build/legs/right` exists.")
        self.assertIsNotNone(right_node)
        left_control = self.comp_layer.lookup('/dummy/build/legs/left/create/fk/controls/upper/create')
        right_control = self.comp_layer.lookup('/dummy/build/legs/right/create/fk/controls/upper/create')
        print("Testing the node "
              "`/dummy/build/legs/left/create/fk/controls/upper/create` "
              "exists.")
        self.assertIsNotNone(left_control)
        print("Testing the node "
              "`/dummy/build/legs/right/create/fk/controls/upper/create` "
              "exists.")
        self.assertIsNotNone(right_control)

    def test_instance_attr_names(self):
        sibling_inst = self.comp_layer.lookup('/Character/build/legs/right')
        inst_attrs = self.stage.get_node_instanced_attr_names(sibling_inst,
                                                              self.comp_layer)
        expected = ['LOCAL']
        self.assertEqual(expected, inst_attrs)


class StageInstance4(unittest.TestCase):

    def setUp(self):
        self.stage = Session().load_file(filepath="./StageRelativeInstanceChildOrder.nxt")
        self.comp_layer = self.stage.build_stage()

    def test_child_orders(self):
        print("Testing that child order remains the same across relative "
              "(sibling) instances")
        expected = ['init', 'build']
        case1_node_path = '/rig_components/limb'
        case1_node = self.comp_layer.lookup(case1_node_path)
        case1_co = self.stage.get_node_child_order(case1_node)
        self.assertEqual(expected, case1_co)

        case2_node_path = '/rig_components/arm'
        case2_node = self.comp_layer.lookup(case2_node_path)
        case2_co = self.stage.get_node_child_order(case2_node)
        self.assertEqual(expected, case2_co)

        case3_node_path = '/rig_components/leg'
        case3_node = self.comp_layer.lookup(case3_node_path)
        case3_co = self.stage.get_node_child_order(case3_node)
        self.assertEqual(expected, case3_co)

    def test_child_order_propagation(self):
        print("Testing that child order propagates to relative (sibling) "
              "instances")
        # Set new child order
        expected = ['build', 'init']
        base_node_path = '/rig_components/_component'
        target = self.stage._sub_layers[0]
        base_node = target.lookup(base_node_path)
        self.stage.set_node_child_order(base_node, expected,
                                        target, comp_layer=self.comp_layer)
        # Test it changed all the instances
        case1_node_path = '/rig_components/limb'
        case1_node = self.comp_layer.lookup(case1_node_path)
        case1_co = self.stage.get_node_child_order(case1_node)
        self.assertEqual(expected, case1_co)

        case2_node_path = '/rig_components/arm'
        case2_node = self.comp_layer.lookup(case2_node_path)
        case2_co = self.stage.get_node_child_order(case2_node)
        self.assertEqual(expected, case2_co)

        case3_node_path = '/rig_components/leg'
        case3_node = self.comp_layer.lookup(case3_node_path)
        case3_co = self.stage.get_node_child_order(case3_node)
        self.assertEqual(expected, case3_co)


class StageRuntimeResolveScenarios(unittest.TestCase):

    def test_run_with_no_change(self):
        """Test Stage Runtime Resolve Scenario1"""
        print("Test SetUp StageRuntimeResolveScenario1")
        os.chdir(os.path.dirname(__file__))
        self.stage = Session().load_file(filepath="./StageRuntimeTest1.nxt")
        self.comp_layer = self.stage.build_stage()
        # Lookup all editor nodes
        self.NodeA = self.comp_layer.lookup('/NodeA')
        self.NodeB = self.comp_layer.lookup('/NodeB')
        # Setup the session layer which generates runtime nodes

        print("Testing runtime attr values...")
        print("Starting execute...")
        self.runtime_layer = self.stage.execute('/NodeA')
        # Lookup all the runtime nodes
        self.rt_NodeA = self.runtime_layer.cache_layer.lookup('/NodeA')
        self.rt_NodeB = self.runtime_layer.cache_layer.lookup('/NodeB')
        print("Test that `/NodeA.attr0` is NOT changed by execute.")
        self.assertEqual(self.NodeA.attr0, self.rt_NodeA.attr0)
        print("Test that `/NodeA.attr0` python type IS changed by execute.")
        self.assertNotEqual(self.NodeA.attr1, self.rt_NodeA.attr1)
        print("Test that `/NodeA.attr0` python type change is as expected.")
        self.assertEqual(int(self.NodeA.attr1), self.rt_NodeA.attr1)

    def test_run_with_change_in_code_a(self):
        print("Test SetUp StageRuntimeResolveScenario2")
        os.chdir(os.path.dirname(__file__))
        self.stage = Session().load_file(filepath="./StageRuntimeTest2.nxt")
        self.comp_layer = self.stage.build_stage()
        # Lookup all editor nodes
        self.NodeA = self.comp_layer.lookup('/NodeA')
        self.NodeB = self.comp_layer.lookup('/NodeB')
        print("Testing runtime attr value mutation by single code block...")
        print("Starting execute...")
        self.runtime_layer = self.stage.execute('/NodeA')
        # Lookup all the runtime nodes again since execute build the session
        # layer again
        self.rt_NodeA = self.runtime_layer.cache_layer.lookup('/NodeA')
        self.rt_NodeB = self.runtime_layer.cache_layer.lookup('/NodeB')
        print("Testing NodeA resolved code code after execute.")
        expected_code = '# printing values before they are changed\n' \
                           'print("side")\n' \
                           'print(5)\n' \
                           'print(self.attr0)\n' \
                           'print(self.attr1)  # notice this value is ' \
                           'the input value!\n' \
                           '\n' \
                           '# changing value from 5 to 10\n' \
                           'self.attr1 = 10\n' \
                           '\n# printing after value change' \
                           '\nprint(self.attr1)  # notice this value change ' \
                           'is a result of runtime\n' \
                           '\n' \
                           '# caveat!  ' \
                           'This is expected behavior - DO NOT modify at ' \
                           'runtime!' \
                           '\n' \
                           'print(5)\n'
        self.assertEqual(expected_code, getattr(self.rt_NodeA, INTERNAL_ATTRS.CACHED_CODE))
        print("Testing NodeA local attributes after execute.")
        self.assertEqual("side", getattr(self.rt_NodeA, 'attr0'))
        self.assertEqual(10, getattr(self.rt_NodeA, 'attr1'))

        print("Testing NodeB resolved code code after execute.")
        expected_code = '# just printing values for now\n' \
                           'print(10)  # this is the correct answer ' \
                           'currently! NodeA hasn\'t run yet!\n' \
                           'print(self.attr2)\n' \
                           '\n' \
                           '# also printing remote attrs for fun!\n' \
                           'print("side")\n' \
                           'print(10)  # this is the correct answer too!'
        self.assertEqual(expected_code, getattr(self.rt_NodeB, INTERNAL_ATTRS.CACHED_CODE))
        print("Testing NodeB local attributes after execute.")
        self.assertEqual(10, getattr(self.rt_NodeB, 'attr2'))

    def test_runtime_prop(self):
        os.chdir(os.path.dirname(__file__))
        stage = Session().load_file(filepath="RuntimeProp.nxt")
        runtime_layer = stage.execute('/inst_src')
        # Check target node, who we expect to have inherited
        # changed values from source node.
        cached_tgt = runtime_layer.cache_layer.lookup('/inst_tgt')
        expected_refer = "\'new\'"
        cached_refer = cached_tgt.refer
        self.assertEqual(expected_refer, cached_refer)
        expected_list = ['old', 'me_first', 'again', 'again']
        cached_list = cached_tgt.list
        self.assertEqual(expected_list, cached_list)
        expected_deep = [[1, 2], ['old']]
        cached_deep = cached_tgt.deep
        self.assertEqual(expected_deep, cached_deep)
        expected_changer = 'me_second'
        cached_changer = cached_tgt.changer
        self.assertEqual(expected_changer, cached_changer)
        expected_changed = 'second'
        cached_changed = cached_tgt.changed
        self.assertEqual(expected_changed, cached_changed)

    def test_future_tokens(self):
        os.chdir(os.path.dirname(__file__))
        # This graph has asserts, if it errors this test fails.
        Session().execute_graph("future_tokens_test.nxt")

    def test_exit_exceptions(self):
        """Test the various gracefully exit exceptions"""
        from nxt.runtime import ExitGraph
        print("Test ExitNode and ExitGraph exceptions")
        os.chdir(os.path.dirname(__file__))
        self.stage = Session().load_file(filepath="./StageRuntimeTest2.nxt")
        self.comp_layer = self.stage.build_stage()

        print("Starting execute...")
        with self.assertRaises(ExitGraph) as exit_graph:
            self.stage.execute('/Start')
        self.runtime_layer = exit_graph.exception.runtime_layer
        # Look up the nodes on the cache layer
        self.start = self.runtime_layer.cache_layer.lookup('/Start')
        self.mid = self.runtime_layer.cache_layer.lookup('/Start/mid')
        self.low = self.runtime_layer.cache_layer.lookup('/Start/mid/low')
        # Never run shouldn't be on the cache layer
        self.never_run = self.runtime_layer.cache_layer.lookup('/Start/never_run')
        print("Testing /Start ran as expected")
        self.assertEqual(True, getattr(self.start, 'i_ran'))
        print("Testing /Start/mid ran as expected")
        self.assertEqual(True, getattr(self.mid, 'i_exited'))
        print("Testing /Start/mid/low ran as expected")
        self.assertEqual(True, getattr(self.low, 'exited_graph'))
        print("Testing /Start/never_run never ran")
        self.assertEqual(None, self.never_run)

    def test_sub_graph_exit(self):
        """Test the various gracefully exit exceptions work in sub-graphs"""
        from nxt.runtime import ExitGraph
        print("Test ExitNode and ExitGraph exceptions")
        os.chdir(os.path.dirname(__file__))
        self.stage = Session().load_file(filepath="./test_sub_graph_exit.nxt")
        self.comp_layer = self.stage.build_stage()

        print("Starting execute...")
        self.runtime_layer = self.stage.execute('/test_1')
        cached_node = self.runtime_layer.cache_layer.lookup('/test_1')
        print('Testing that graph exited and pushed its STAGE values up...')
        self.assertFalse(getattr(cached_node, 'bad', True))

    def test_python_sets(self):
        self.stage = Session().load_file(filepath="./StageRuntimeTest2.nxt")
        self.comp_layer = self.stage.build_stage()
        self.runtime_layer = self.stage.execute('/set_test0', self.comp_layer)
        expected = set(range(10))
        actual = self.stage.get_attr_as_real_data_type(self.runtime_layer.lookup('/set_test0'), 'a', self.runtime_layer)
        print('Test start node set is correct')
        self.assertSetEqual(expected, actual)
        actual = self.stage.get_attr_as_real_data_type(self.runtime_layer.lookup('/set_test1'), 'b', self.runtime_layer)
        print('Test end node set is correct')
        self.assertSetEqual(expected, actual)


class StageChildOrder(unittest.TestCase):
    """Unit test relies on the following save files:
    ./StageChildOrderTest_TopLayer.nxt
    ./StageChildOrderTest.nxt"""

    def setUp(self):
        """Opens ./StageChildOrderTest_TopLayer.nxt which has the sub-layer
        ./StageChildOrderTest.nxt"""
        os.chdir(os.path.dirname(__file__))
        self.stage = Session().load_file(filepath="./StageChildOrderTest_TopLayer.nxt")
        self.tgt_layer = self.stage._sub_layers[0]
        self.lower_tgt_layer = self.stage._sub_layers[1]
        self.comp_layer = self.stage.build_stage()
        self.inst_tgt_path = '/InstanceTarget'
        self.parent_path = '/ParentNode'
        self.lower_parent_path = '/LowerParent'
        self.tgt_path = '/tgt'
        self.parent_node = self.comp_layer.lookup(self.parent_path)
        self.instance_target_node = self.comp_layer.lookup(self.inst_tgt_path)
        self.lower_parent_node = self.comp_layer.lookup(self.lower_parent_path)

    def test_get_child_order(self):
        """Test getting a node's child order via a get attr and the stages
        method"""
        expected_child_order = ['node0', 'node1', 'node2']
        print("Testing get child order on a real node with the same namespace "
              "as a lower layer node. "
              "Expected child order: {}".format(expected_child_order))
        actual_child_order = getattr(self.parent_node,
                                     INTERNAL_ATTRS.CHILD_ORDER)
        self.assertEqual(expected_child_order, actual_child_order)
        self.assertEqual(expected_child_order,
                         self.stage.get_node_child_order(self.parent_node))
        expected_child_order = ['Child0', 'Child1']
        print("Testing get child order on an instance source node. "
              "Expected child order: {}".format(expected_child_order))
        actual_child_order = getattr(self.lower_parent_node,
                                     INTERNAL_ATTRS.CHILD_ORDER)
        self.assertEqual(expected_child_order, actual_child_order)
        self.assertEqual(expected_child_order,
                         self.stage.get_node_child_order(
                             self.lower_parent_node))
        expected_child_order = ['Child2', 'Child0', 'Child1']
        print("Testing get child order on an instance target node. "
              "Expected child order: {}".format(expected_child_order))
        actual_child_order = getattr(self.instance_target_node,
                                     INTERNAL_ATTRS.CHILD_ORDER)
        self.assertEqual(expected_child_order, actual_child_order)
        self.assertEqual(expected_child_order,
                         self.stage.get_node_child_order(
                             self.instance_target_node))
        children = self.comp_layer.children(self.inst_tgt_path, ordered=True)
        children_names = [getattr(n, INTERNAL_ATTRS.NAME) for n in children]
        expected_children_names = ['Child2', 'Child0', 'Child1']
        print("Testing children names. "
              "Expected children: {}".format(expected_children_names))
        self.assertEqual(expected_children_names, children_names)

    def test_set_child_order_basic(self):
        """Test setting a node's child order via stage's method. Validated
        via a getattr and the stage's method"""
        expected_child_order = ['node1', 'node2', 'node0']
        print("Testing set child order on a real node with the same namespace "
              "as a lower layer node. "
              "Expected child order: {}".format(expected_child_order))
        parent_node = self.tgt_layer.lookup(self.parent_path)
        self.stage.set_node_child_order(parent_node,
                                        expected_child_order, self.tgt_layer)
        # Rebuild stage and look up nodes
        self.comp_layer = self.stage.build_stage()
        self.parent_node = self.comp_layer.lookup(self.parent_path)
        self.instance_target_node = self.comp_layer.lookup(self.inst_tgt_path)
        self.lower_parent_node = self.comp_layer.lookup(self.lower_parent_path)
        # Test child order
        actual_child_order = getattr(self.parent_node,
                                     INTERNAL_ATTRS.CHILD_ORDER)
        self.assertEqual(expected_child_order, actual_child_order)
        self.assertEqual(expected_child_order,
                         self.stage.get_node_child_order(self.parent_node))
        expected_child_order = ['Child1', 'Child0']
        print("Testing set child order on an instance source node. "
              "Expected child order: {}".format(expected_child_order))
        lower_parent_node = self.lower_tgt_layer.lookup(self.lower_parent_path)
        self.stage.set_node_child_order(lower_parent_node,
                                        expected_child_order,
                                        self.lower_tgt_layer)
        # Rebuild stage and look up nodes
        self.comp_layer = self.stage.build_stage()
        self.parent_node = self.comp_layer.lookup(self.parent_path)
        self.instance_target_node = self.comp_layer.lookup(self.inst_tgt_path)
        self.lower_parent_node = self.comp_layer.lookup(self.lower_parent_path)
        actual_child_order = getattr(self.lower_parent_node,
                                     INTERNAL_ATTRS.CHILD_ORDER)
        self.assertEqual(expected_child_order, actual_child_order)
        self.assertEqual(expected_child_order,
                         self.stage.get_node_child_order(self.lower_parent_node))

    def test_set_child_order_with_propagate(self):
        """Tests that child order is properly propagated through instances."""
        expected_child_order = ['Child1', 'Child0']
        lower_parent_node = self.lower_tgt_layer.lookup(self.lower_parent_path)
        self.stage.set_node_child_order(lower_parent_node,
                                        expected_child_order,
                                        self.lower_tgt_layer, self.comp_layer)
        # Rebuild stage and look up nodes
        self.comp_layer = self.stage.build_stage()
        self.parent_node = self.comp_layer.lookup(self.parent_path)
        self.instance_target_node = self.comp_layer.lookup(self.inst_tgt_path)
        self.lower_parent_node = self.comp_layer.lookup(self.lower_parent_path)
        # Test child order
        actual_child_order = getattr(self.lower_parent_node,
                                     INTERNAL_ATTRS.CHILD_ORDER)
        self.assertEqual(expected_child_order, actual_child_order)
        self.assertEqual(expected_child_order,
                         self.stage.get_node_child_order(
                             self.lower_parent_node))
        expected_child_order = ['Child2', 'Child1', 'Child0']
        print("Testing set child order propagation. "
              "Expected child order: {}".format(expected_child_order))
        actual_child_order = getattr(self.instance_target_node,
                                     INTERNAL_ATTRS.CHILD_ORDER)
        self.assertEqual(expected_child_order, actual_child_order)
        self.assertEqual(expected_child_order,
                         self.stage.get_node_child_order(
                             self.instance_target_node))

    def test_child_order_with_add_node(self):
        expected_child_order = ['Child1', 'Child2', 'Child3']
        print("Testing localizing a node does not break child order. "
              "Expected child order: {}".format(expected_child_order))
        tgt_node = self.comp_layer.lookup(self.tgt_path)
        tgt_node_spec = self.tgt_layer.lookup(self.tgt_path)
        self.assertEqual(expected_child_order,
                         getattr(tgt_node, INTERNAL_ATTRS.CHILD_ORDER))
        # Localize child 1
        self.stage.add_node(name='Child1', data=None, parent=tgt_node_spec,
                            layer=self.tgt_layer, fix_names=False,
                            comp_layer=self.comp_layer)
        self.assertEqual(expected_child_order,
                         getattr(tgt_node, INTERNAL_ATTRS.CHILD_ORDER))
        self.stage.delete_node(self.tgt_layer.lookup(self.tgt_path + '/Child1'),
                               self.tgt_layer, self.comp_layer)
        self.assertEqual(expected_child_order,
                         getattr(tgt_node, INTERNAL_ATTRS.CHILD_ORDER))
        # Localize child 2
        self.stage.add_node(name='Child2', data=None, parent=tgt_node_spec,
                            layer=self.tgt_layer, fix_names=False,
                            comp_layer=self.comp_layer)
        self.assertEqual(expected_child_order,
                         getattr(tgt_node, INTERNAL_ATTRS.CHILD_ORDER))
        self.stage.delete_node(self.tgt_layer.lookup(self.tgt_path+'/Child2'),
                               self.tgt_layer, self.comp_layer)
        self.assertEqual(expected_child_order,
                         getattr(tgt_node, INTERNAL_ATTRS.CHILD_ORDER))
        # Localize child 3
        self.stage.add_node(name='Child3', data=None, parent=tgt_node_spec,
                            layer=self.tgt_layer, fix_names=False,
                            comp_layer=self.comp_layer)
        self.assertEqual(expected_child_order,
                         getattr(tgt_node, INTERNAL_ATTRS.CHILD_ORDER))
        self.stage.delete_node(self.tgt_layer.lookup(self.tgt_path + '/Child3'),
                               self.tgt_layer, self.comp_layer)
        self.assertEqual(expected_child_order,
                         getattr(tgt_node, INTERNAL_ATTRS.CHILD_ORDER))
        # New node
        expected_child_order = expected_child_order + ['Bob']
        print("Testing a new node goes to the top. "
              "Expected child order: {}".format(expected_child_order))
        self.stage.add_node(name='Bob', data=None, parent=tgt_node_spec,
                            layer=self.tgt_layer, fix_names=False,
                            comp_layer=self.comp_layer)
        self.assertEqual(expected_child_order,
                         getattr(tgt_node, INTERNAL_ATTRS.CHILD_ORDER))
        self.stage.delete_node(self.tgt_layer.lookup(self.tgt_path + '/Bob'),
                               self.tgt_layer, self.comp_layer)

    def test_child_order_merger(self):
        print("Test that the child order merger works in edge cases.")
        src = ['node2', 'node']
        tgt = ['node1', 'node2']
        expected = ['node1', 'node2', 'node']
        print("Test that merging {} <-> {} becomes {}".format(src, tgt,
                                                              expected))
        result = nxt_node.list_merger(src, tgt)
        self.assertEqual(expected, result)


class StageChildProxyOrder(unittest.TestCase):
    """Unit test relies on the following save files:
    ./test_proxy_stack.nxt
    """

    def setUp(self):
        """Opens ./StageChildOrderTest_TopLayer.nxt which has the sub-layer
        ./StageChildOrderTest.nxt"""
        os.chdir(os.path.dirname(__file__))
        self.stage = Session().load_file(filepath="./test_proxy_stack.nxt")
        self.tgt_layer = self.stage._sub_layers[0]
        self.comp_layer = self.stage.build_stage()
        self.parent_path = '/instance2/first_child/bot_gran'
        self.parent_node = self.comp_layer.lookup(self.parent_path)

    def test_get_child_order(self):
        """Test getting a node's child order via a get attr and the stages
        method"""
        expected_child_order = ['/instance2/first_child/bot_gran/bot_great_gran']
        print("Testing get child order on a proxy node with a real child."
              "Expected child order: {}".format(expected_child_order))
        actual_child_order = getattr(self.parent_node,
                                     INTERNAL_ATTRS.CHILD_ORDER)
        self.assertEqual([], actual_child_order)
        self.assertEqual(expected_child_order,
                         self.comp_layer.children(self.parent_path,
                                                  return_type=self.comp_layer.RETURNS.Path))


class StageAddNode(unittest.TestCase):
    def test_basic_name_collision(self):
        """Verify when adding nodes without specific names to a layer they
        will not collide."""
        test_stage = stage.Stage()
        node_count = 3
        for _ in range(node_count):
            test_stage.add_node()
        comp_layer = test_stage.build_stage()
        all_names = [getattr(c, INTERNAL_ATTRS.NAME) for c in
                     comp_layer.children()]
        all_names = set(all_names)
        self.assertEqual(node_count, len(all_names))

    def test_upper_respects_lower(self):
        """Verify that when an add node command is given targeting an upper layer,
        an override of a lower layer will not be created.
        NOTE: this test depends on the test graphs `UpperLowerRespect.nxt` and
        `LowerRespect.nxt`."""
        os.chdir(os.path.dirname(__file__))
        test_stage = Session().load_file(filepath='UpperLowerRespect.nxt')
        # "lower" is the name of a node present in "LowerRespect.nxt" which
        # is referenced into "UpperRespect.nxt".
        test_stage.add_node('lower', layer=0)
        comp_layer = test_stage.build_stage()
        # There is exactly 1 node in Lower, and 0 in upper. If the above add
        # worked as expected, we should have 2.
        self.assertEqual(2, len(comp_layer.children()))


class StageRuntimeScope(unittest.TestCase):
    def setUp(self):
        os.chdir(os.path.dirname(__file__))
        self.stage = Session().load_file(filepath='./StageRuntimeScope.nxt')

    def test_global_scope(self):
        """Test executes a node that sets "math" as a global and imports math.
        Next another node tries to access the global "math" by running
        math.floor(1.5)"""
        print("Testing that globals are accessible by down stream nodes.")
        comp = self.stage.build_stage()
        runtime_layer = self.stage.execute_nodes(['/setup', '/g'], comp)
        global_node = runtime_layer.lookup('/g')
        self.assertEqual(1.0, global_node.result)

    def test_lambda_scope(self):
        """Executes a node that tests a lambda that used to fail for Zach."""
        print("Testing lambda function that used to fail for Zach.")
        runtime_layer = self.stage.execute(start='/setup')
        lambda_node = runtime_layer.lookup('/lam')
        self.assertEqual('l', lambda_node.result)


class StageNamespaceMerger(unittest.TestCase):
    """Based on real world test cases we test the the namespace merger does
    what is expected."""
    def test_namespace_case01(self):
        source = ['Character', 'build', 'legs', 'left']
        target = ['dummy', 'build', 'legs']
        expected = ['dummy', 'build', 'legs', 'left']
        print("Source: {0}\nTarget: {1}\nExpected: {2}".format(source,
                                                                target,
                                                                expected))
        inst_ns = stage.Stage.namespace_merger(source, target)
        self.assertEqual(expected, inst_ns)

    def test_namespace_case02(self):
        source = ['Character']
        target = ['dummy']
        expected = ['dummy']
        print("Source: {0}\nTarget: {1}\nExpected: {2}".format(source,
                                                                target,
                                                                expected))
        inst_ns = stage.Stage.namespace_merger(source, target)
        self.assertEqual(expected, inst_ns)

    def test_namespace_case03(self):
        source = ['Character', 'build', 'legs', 'left']
        target = ['dummy', 'build', 'legs', 'jack']
        expected = ['dummy', 'build', 'legs', 'left']
        print("Source: {0}\nTarget: {1}\nExpected: {2}".format(source,
                                                                target,
                                                                expected))
        inst_ns = stage.Stage.namespace_merger(source, target)
        self.assertEqual(expected, inst_ns)

    def test_namespace_case04(self):
        source = ['Character', 'build', 'legs', 'left']
        target = ['Character', 'build', 'legs', 'right']
        expected = ['Character', 'build', 'legs', 'left']
        print("Source: {0}\nTarget: {1}\nExpected: {2}".format(source,
                                                                target,
                                                                expected))
        inst_ns = stage.Stage.namespace_merger(source, target)
        self.assertEqual(expected, inst_ns)

    def test_namespace_case05(self):
        source = ['Character', 'build']
        target = ['dummy']
        expected = ['dummy', 'build']
        print("Source: {0}\nTarget: {1}\nExpected: {2}".format(source,
                                                                target,
                                                                expected))
        inst_ns = stage.Stage.namespace_merger(source, target)
        self.assertEqual(expected, inst_ns)

    def test_namespace_case06(self):
        source = ['control', 'create']
        target = ['leg', 'create', 'fk', 'controls', 'upper']
        expected = ['leg', 'create', 'fk', 'controls', 'upper', 'create']
        print("Source: {0}\nTarget: {1}\nExpected: {2}".format(source,
                                                                target,
                                                                expected))
        inst_ns = stage.Stage.namespace_merger(source, target)
        self.assertEqual(expected, inst_ns)

    def test_namespace_case07(self):
        source = ['leg', 'create', 'fk', 'controls']
        target = ['Character', 'build', 'legs', 'left', 'create', 'create',
                  'fk']
        expected = ['Character', 'build', 'legs', 'left', 'create',
                          'create', 'fk', 'controls']
        print("Source: {0}\nTarget: {1}\nExpected: {2}".format(source,
                                                                target,
                                                                expected))
        inst_ns = stage.Stage.namespace_merger(source, target)
        self.assertEqual(expected, inst_ns)

    def test_namespace_case08(self):
        source = ['Character', 'build', 'legs']
        target = ['dummy', 'build']
        expected = ['dummy', 'build', 'legs']
        print("Source: {0}\nTarget: {1}\nExpected: {2}".format(source,
                                                                target,
                                                                expected))
        inst_ns = stage.Stage.namespace_merger(source, target)
        self.assertEqual(expected, inst_ns)

    def test_namespace_case09(self):
        source = ['Character', 'build', 'legs', 'left', 'create']
        target = ['Character', 'build', 'legs', 'right']
        expected = ['Character', 'build', 'legs', 'right', 'create']
        print("Source: {0}\nTarget: {1}\nExpected: {2}".format(source,
                                                                target,
                                                                expected))
        inst_ns = stage.Stage.namespace_merger(source, target)
        self.assertEqual(expected, inst_ns)

    def test_namespace_case10(self):
        source = ['limb', 'create', 'fk']
        target = ['arm', 'limb', 'create']
        expected = ['arm', 'limb', 'create', 'fk']
        print("Source: {0}\nTarget: {1}\nExpected: {2}".format(source,
                                                                target,
                                                                expected))
        inst_ns = stage.Stage.namespace_merger(source, target)
        self.assertEqual(expected, inst_ns)

    def test_namespace_case11(self):
        source = ['spline', 'controls', 'hips']
        target = ['center', 'spine', 'controls']
        expected = ['center', 'spine', 'controls', 'hips']
        print("Source: {0}\nTarget: {1}\nExpected: {2}".format(source,
                                                                target,
                                                                expected))
        inst_ns = stage.Stage.namespace_merger(source, target)
        self.assertEqual(expected, inst_ns)

    def test_namespace_case12(self):
        source = ['arm', 'limb']
        target = ['left', 'arm']
        expected = ['left', 'arm', 'limb']
        print("Source: {0}\nTarget: {1}\nExpected: {2}".format(source,
                                                                target,
                                                                expected))
        inst_ns = stage.Stage.namespace_merger(source, target)
        self.assertEqual(expected, inst_ns)

    def test_namespace_case13(self):
        source = ['arm', 'clavicle', 'control']
        target = ['leg', 'pelvis']
        expected = ['leg', 'pelvis', 'control']
        print("Source: {0}\nTarget: {1}\nExpected: {2}".format(source,
                                                                target,
                                                                expected))
        inst_ns = stage.Stage.namespace_merger(source, target)
        self.assertEqual(expected, inst_ns)

    def test_namespace_case14(self):
        source = ['spine', 'node1']
        target = ['center', 'spine']
        expected = ['center', 'spine', 'node1']
        print("Source: {0}\nTarget: {1}\nExpected: {2}".format(source,
                                                                target,
                                                                expected))
        inst_ns = stage.Stage.namespace_merger(source, target)
        self.assertEqual(expected, inst_ns)


class TestExecOrder(unittest.TestCase):
    def setUp(self):
        os.chdir(os.path.dirname(__file__))
        self.stage = Session().load_file(filepath='order.nxt')
        self.comp_layer = self.stage.build_stage()

    def test_roots_order(self):
        expected = ['/one', '/seven', '/twelve']
        found = self.comp_layer.get_root_exec_order('/one')
        self.assertEqual(expected, found)

        expected = ['/seven', '/twelve']
        found = self.comp_layer.get_root_exec_order('/seven')
        self.assertEqual(expected, found)

    def test_full_order(self):
        expected = ['/one', '/one/two', '/one/three', '/one/three/four',
                    '/one/three/five', '/one/six', '/seven', '/seven/eight',
                    '/seven/eight/nine', '/seven/eight/nine/ten',
                    '/seven/eight/eleven', '/twelve']
        found = self.comp_layer.get_exec_order('/one')
        self.assertEqual(expected, found)

        expected = ['/one/three/five', '/one/six', '/seven', '/seven/eight',
                    '/seven/eight/nine', '/seven/eight/nine/ten',
                    '/seven/eight/eleven', '/twelve']
        found = self.comp_layer.get_exec_order('/one/three/five')
        self.assertEqual(expected, found)

    def test_with_disabled(self):
        found = self.comp_layer.get_exec_order('/a')
        expected = ['/a', '/a/b', '/c']
        self.assertEqual(expected, found)

    def test_under_disabled_parent(self):
        found = self.comp_layer.get_exec_order('/SKIP/NO')
        expected = ['/SKIP/NO', '/SKIP/MOON', '/c']
        self.assertEqual(expected, found)


class TestExecOrder2(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.stage = Session().load_file(filepath="./StageInstanceTest_Layer0.nxt")
        cls.comp_layer = cls.stage.build_stage()

    def test_child_cache(self):
        found = self.comp_layer.get_exec_order('/exec_test')
        expected = ['/exec_test', '/exec_test/exec_test_child',
                    '/exec_test/exec_test_child/could_be_missing']
        self.assertEqual(expected, found)


class TestFutureImport(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.stage = Session().load_file(filepath="./StageRuntimeTest1.nxt")
        cls.comp_layer = cls.stage.build_stage()

    def test_child_cache(self):
        print("Test that `from __future__ import` works without error")
        layer = self.stage.execute_nodes(['/NodeC'], self.comp_layer)
        node = layer.lookup('/')
        passed = stage.get_node_attr(node, 'passed', False)
        self.assertTrue(passed)


if __name__ == '__main__':
    unittest.main()



# Built-in
import os
import json

version_file = os.path.join(os.path.dirname(__file__), 'version.json')


class API_VERSION(object):
    with open(version_file, 'r') as f:
        version_data = json.load(f)
    api_v_data = version_data['API']
    MAJOR = api_v_data['MAJOR']
    MINOR = api_v_data['MINOR']
    PATCH = api_v_data['PATCH']
    VERSION_TUPLE = (MAJOR, MINOR, PATCH)
    VERSION_STR = '.'.join(str(v) for v in VERSION_TUPLE)
    VERSION = VERSION_STR


class GRAPH_VERSION(object):
    with open(version_file, 'r') as f:
        version_data = json.load(f)
    api_v_data = version_data['GRAPH']
    MAJOR = api_v_data['MAJOR']
    MINOR = api_v_data['MINOR']
    VERSION_TUPLE = (MAJOR, MINOR)
    VERSION_STR = '.'.join(str(v) for v in VERSION_TUPLE)
    VERSION = VERSION_STR


# User and Site Dirs
USER_DIR_ENV_VAR = 'NXT_USER_DIR'
SITE_DIR_ENV_VAR = 'NXT_SITE_DIR'

if USER_DIR_ENV_VAR in os.environ:
    USER_DIR = os.environ.get(USER_DIR_ENV_VAR)
else:
    USER_DIR = os.path.expanduser(os.path.join('~', 'nxt'))

SITE_DIR = USER_DIR
if SITE_DIR_ENV_VAR in os.environ:
    SITE_DIR = os.environ.get(SITE_DIR_ENV_VAR)
# Configs
CONFIG_VERSION = str(API_VERSION.MAJOR)
USER_CONFIG_DIR = os.path.join(USER_DIR, 'config', CONFIG_VERSION)
if not os.path.isdir(USER_CONFIG_DIR):
    os.makedirs(USER_CONFIG_DIR)
SITE_CONFIG_DIR = os.path.join(SITE_DIR, 'config', CONFIG_VERSION)
# Plugins
PLUGIN_DIR_NAME = 'plugins'
BUILTIN_PLUGIN_DIR = os.path.join(os.path.dirname(__file__), PLUGIN_DIR_NAME)
USER_PLUGIN_DIR = os.path.join(USER_CONFIG_DIR, PLUGIN_DIR_NAME)
if not os.path.isdir(USER_PLUGIN_DIR):
    os.makedirs(USER_PLUGIN_DIR)
SITE_PLUGIN_DIR = os.path.join(SITE_CONFIG_DIR, PLUGIN_DIR_NAME)
PLUGIN_DIRS = [USER_PLUGIN_DIR, SITE_PLUGIN_DIR, BUILTIN_PLUGIN_DIR]


class DATA_STATE:
    RAW = 'raw'
    RESOLVED = 'resolved'
    CACHED = 'cached'


class NODE_ERRORS:
    INSTANCE = 'Broken instance'
    EXEC = 'Broken exec in'
    ORPHANS = 'Orphan children'


class FILE_FORMAT(object):
    ASCII = '.nxt'
    BINARY = '.nxtb'
    _ALL = (ASCII, BINARY)


UNTITLED = 'untitled'
IGNORE = '<!ignore!>'
GRID_SIZE = 20  # Must be int
NXT_DCC_ENV_VAR = 'NXT_DCC'
STANDALONE = 'standalone'


def is_standalone():
    return os.environ.get(NXT_DCC_ENV_VAR) == STANDALONE


# Builtin
import os
import tempfile
import logging

logger = logging.getLogger('nxt')


def get_running_server_address(as_str=True):
    host = 'localhost'
    port = 8001
    server_info_filepath = get_server_info_filepath()
    valid_data = False
    if os.path.isfile(server_info_filepath):
        with open(server_info_filepath, 'r') as fp:
            lines = fp.readlines()
            if len(lines) == 2:
                _, port = lines
                port = int(port)
                valid_data = True
    if not valid_data:
        logger.error('No running server could be found, returning defaults!')
    if as_str:
        return '{}:{}'.format(host, port)
    else:
        return host, port


def get_server_info_filepath():
    temp_dir = tempfile.gettempdir()
    server_info_filepath = os.path.join(temp_dir, 'NXT_RPC_SERVER')
    return server_info_filepath



import os
import unittest

from nxt.session import Session


class SubGraphs(unittest.TestCase):
    def setUp(self):
        os.chdir(os.path.dirname(__file__))
        self.stage = Session().load_file(filepath="test_sub_graph.nxt")

    def test_basic_run(self):
        """Test that a sub graph is run. The node that is run calls a sub-graph
        that creates a file. If that file is there, we can assert the garph has
        run.
        """
        expected_file = os.path.join(os.path.dirname(__file__),
                                     'sub_graph_file.txt')
        # The file must not exist if we are going to believe the graph made it.
        if os.path.isfile(expected_file):
            os.remove(expected_file)
        self.assertFalse(os.path.isfile(expected_file))
        self.stage.execute(start='/basic_file')
        self.assertTrue(os.path.isfile(expected_file))
        # Clean up after the test
        os.remove(expected_file)

    def test_parameters(self):
        """Test that local attributes of a node that calls a sub graph are
        pushed to the sub graph world node, and that resulting attributes of
        the sub graph world node are pulled back into the calling node.
        """
        target_path = '/specific_file'
        comp_layer = self.stage.build_stage()
        specific_node = comp_layer.lookup(target_path)
        expected_file = self.stage.get_node_attr_value(specific_node,
                                                       'location', comp_layer)
        # The file must not exist if we are going to believe the graph made it.
        if os.path.isfile(expected_file):
            os.remove(expected_file)
        self.assertFalse(os.path.isfile(expected_file))
        self.stage.execute(start=target_path)
        self.assertTrue(os.path.isfile(expected_file))
        # Clean up after the test
        os.remove(expected_file)



# Built-in
import argparse
import sys
import os
import json
import logging
import time

# Internal
from nxt.session import Session

from nxt import legacy
from nxt import nxt_log
from nxt.constants import (API_VERSION, GRAPH_VERSION, NXT_DCC_ENV_VAR,
                           STANDALONE)
from nxt.remote.contexts import iter_context_names
has_editor = False
try:
    import nxt_editor
    from nxt_editor.constants import EDITOR_VERSION
    has_editor = True
except ImportError:
    pass

logger = logging.getLogger('nxt')


class InvalidChoice(Exception):
    pass


class UnrecognizedArg(Exception):
    pass


class CustomParser(argparse.ArgumentParser):
    def error(self, message):
        if "invalid choice" in message:
            # If message doesn't look like a file path, raise the error
            if '.nxt' not in message:
                super(CustomParser, self).error(message)
            raise InvalidChoice(message)
        elif 'unrecognized arguments' in message:
            possible_args = ['-gui']
            args = [a for a in possible_args if a in message]
            if args:
                raise UnrecognizedArg(message)

        super(CustomParser, self).error(message)


def count_down(t=3):
    while t != 0:
        sys.stdout.write('\r' + str(t))
        sys.stdout.flush()
        time.sleep(.8)
        t -= 1
    sys.stdout.write('\r ')
    sys.stdout.flush()


def get_version():
    # Get API version
    api_v = API_VERSION.VERSION_STR
    # Get editor version
    try:
        editor_v = EDITOR_VERSION.VERSION_STR
    except NameError:
        editor_v = False
    # Get graph version
    graph_v = GRAPH_VERSION.VERSION_STR
    # Assemble version string
    if editor_v:
        version_str = 'API {av} | Editor {ev} |  Graph {gv}'
    else:
        version_str = 'API {av} | Graph {gv}'
    return version_str.format(av=api_v, ev=editor_v, gv=graph_v)


def editor(args):
    """Launches editor

    :param args: Namespace Object
    :return: None
    """
    if not has_editor:
        msg = 'Editor not found, you can install with "pip install nxt_editor"'
        print(msg)
        return
    if isinstance(args.path, list):
        paths = args.path
    else:
        paths = [args.path]
    os.environ[NXT_DCC_ENV_VAR] = STANDALONE
    sys.exit(nxt_editor.launch_editor(paths, start_rpc=not args.no_rpc))


def execute(args):
    """Executes graph

    :param args: Namespace Object
    :return: None"""
    if not hasattr(args, 'parameters'):
        parameter_list = []  # Legacy does not support parameters
    else:
        parameter_list = args.parameters
    if not hasattr(args, 'start'):
        start = None # Legacy does not support start points
    else:
        start = args.start

    param_arg_count = len(parameter_list)
    if param_arg_count == 1:
        # Read file for parameters
        param_path = parameter_list[0]
        if not os.path.isfile(param_path):
            msg = 'Single parameter passed, expected it to be valid parameters file. However, "{}" does not exist'.format(param_path)
            raise IOError(msg)
        with open(param_path, 'r') as fp:
            parameters = json.load(fp)
    elif param_arg_count % 2 != 0:
        raise Exception('Invalid parameters supplied, must be 1 file path or '
                        'pattern: "/node.attr value"')
    else:
        # Parse cli formatted input
        parameters = {}
        i = 0
        for _ in range(int(param_arg_count / 2)):
            key = parameter_list[i]
            if not key.startswith('/'):
                raise Exception('Invalid attr path key {}, must be '
                                'formatted as /node.attr'.format(key))
            val = parameter_list[i + 1]
            parameters[key] = val
            i += 2
    context = getattr(args, 'context', None)
    Session().execute_graph(args.path[0], start, parameters, context)
    logger.execinfo('Execution finished!')


def convert(args):
    """Convert save file

    :param args: Namespace Object
    :return: None"""
    legacy.cli_file_convert(args.path[0], args.replace)
    sys.exit()


def main():
    desc = 'execute nxt file or open an nxt session'
    parser = CustomParser(description=desc, prog='nxt')
    subs = parser.add_subparsers()
    parser.add_argument('-v', '--verbose', help='verbose execution '
                                                '(-vv for debugging)',
                        action='count')
    parser.add_argument('--version', action='version', version=get_version())
    # Legacy
    # TODO: Remove at 2.0
    leg_desc = ('Please convert your calls to the new system: '
                'edit, exec,or convert.')
    legacy_parser = subs.add_parser('legacy', help=leg_desc)
    legacy_parser.add_argument('-v', '--verbose', help='verbose execution',
                               action='store_true')
    if has_editor:
        gui_parser = subs.add_parser('ui', help='Launch visual editor.')
        gui_parser.set_defaults(which='ui')
        gui_parser.add_argument('path', type=str, nargs='?',
                                help='file(s) to open', default='')
        no_rpc_help = ('Start editor without setting up an rpc server during '
                       'startup.')
        gui_parser.add_argument('-no-rpc', help=no_rpc_help,
                                action='store_true')

    exec_parser = subs.add_parser('exec', help='Execute graph. See: exec -h')
    exec_parser.set_defaults(which='exec')

    exec_parser.add_argument('path', type=str, nargs=1, help='file to execute')
    exec_parser.add_argument('-s', '--start', nargs='?', default=None,
                             help='start node path')
    exec_parser.add_argument('-c', '--context', nargs='?', default=None,
                             help='optional remote context to call graph in',
                             choices=list(iter_context_names()))

    convert_parser = subs.add_parser('convert', help='upgrades old save file to'
                                                     ' current version.'
                                                     'See: convert -h')
    convert_parser.set_defaults(which='convert')

    convert_parser.add_argument('path', type=str, nargs=1,
                                help='file/dir to convert')

    convert_parser.add_argument('-r', '--replace', help='replace file with '
                                                        'converted.')

    parameters_help = '''Incompatible with -gui! Specify node attributes to
    overload before running the graph. Specify via several in-line arguments,
    or a single path to json file in the following format.
        /node.attr 5
        /node.third_attr "\\"\\"\\"Hello World!\\"\\"\\""
    Escape literal double quotes, try not to use them, if you
    have to, use the Python or file API.
    '''
    exec_parser.add_argument('-p', '--parameters', nargs="*",
                             help=parameters_help, default=())

    legacy_parser.add_argument('-gui', '--gui',
                               help='launch visual editor session '
                                    '(will be depreciated at 2.0)',
                               action='store_true')
    legacy_parser.add_argument('path', type=str, nargs='?', help='file to open',
                               default='')
    legacy_parser.add_argument('-c', '--convert',
                               help='converts old save version to '
                                    'current (will be depreciated at 2.0)',
                               action='store_true')
    legacy_parser.add_argument('-r', '--replace',
                               help='when used with -c the file on disc is '
                                    'replaced (will be depreciated at 2.0)',
                               action='store_true')
    legacy_parser.add_argument('path', type=str, help='file to open',
                               default='')
    test_parser = subs.add_parser('test', help='Runs unit tests')
    test_parser.set_defaults(which='test')
    if '--' in sys.argv:
        idx = sys.argv.index('--')
        sys.argv = sys.argv[idx:]
    try:
        args = parser.parse_args()
    except (InvalidChoice, UnrecognizedArg):
        # Supports legacy
        logger.exception('Invalid choice, falling back to legacy.')
        try:
            sys.argv.remove('legacy')
        except ValueError:
            pass
        args = legacy_parser.parse_args()
        setattr(args, 'which', 'legacy')
        if getattr(args, 'path'):
            if not isinstance(args.path, list):
                args.path = [args.path]
    if args.verbose:
        if os.environ.get(nxt_log.VERBOSE_ENV_VAR) is None:
            os.environ[nxt_log.VERBOSE_ENV_VAR] = str(args.verbose)
        nxt_log.set_verbosity(args.verbose)
    if not hasattr(args, 'which'):
        parser.parse_args(['-h'])
        return
    if args.which == 'legacy':
        if args.gui:
            logger.warning('The flag -gui will be depreciated at 2.0, please '
                           'see: ui -h')
            count_down()
            editor(args)
        if args.convert:
            logger.warning('The flag -c will be depreciated at 2.0, please '
                           'see: convert -h')
            count_down()
            convert(args)
        else:
            logger.warning('Executing without the "exec" keyword will be '
                           'depreciated at 2.0, please '
                           'see: exec -h')
            count_down()
            execute(args)

    elif args.which == 'convert':
        if not args.path:
            raise IOError('No file path to convert provided!')
        convert(args)
    elif args.which == 'test':
        d = os.path.dirname(__file__)
        path = os.path.join(d, 'test/unittests.nxt').replace(os.sep, '/')
        test_args = argparse.Namespace(path=[path], context=None)
        execute(test_args)
    elif args.which == 'ui':
        editor(args)
    elif args.path and args.which == 'exec':
        execute(args)
    else:
        logger.error('No file path provided!')


if __name__ == '__main__':
    main()



# Builtin
import os

# External
from nxt.remote.contexts import RemoteContext, register_context

# Setup Context
_name = '{name}'
_exe = '{interpreter_exe}'
_graph = '{context_graph}'
_args = {args}
_context = RemoteContext(_name, _exe, _graph, _args)
register_context(_context)



# Built-in
import os
import imp
import logging

# Internal
from .constants import PLUGIN_DIRS

_nxt_loaded_plugin_module_names = []
_nxt_loaded_plugin_modules = []

logger = logging.getLogger('nxt.plugins')


def load_plugins():
    global _nxt_loaded_plugin_module_names
    global _nxt_loaded_plugin_modules
    logger.info('Loading plugins.')
    for plugin_dir in PLUGIN_DIRS:
        if not os.path.isdir(plugin_dir):
            continue
        for file_name in os.listdir(plugin_dir):
            mod_name, _ = os.path.splitext(file_name)
            if mod_name in _nxt_loaded_plugin_module_names:
                continue
            try:
                found = imp.find_module(mod_name, [plugin_dir])
            except ImportError:
                continue
            mod = imp.load_module(mod_name, *found)
            _nxt_loaded_plugin_module_names += [mod_name]
            _nxt_loaded_plugin_modules += [mod]



# Built-in
import unittest
import sys
import os

# Internal
from nxt import nxt_path
from nxt.nxt_node import INTERNAL_ATTRS
from nxt.session import Session


class TestExpansion(unittest.TestCase):
    def test_vars_only(self):
        platform_tests = {
            'linux': "$zoom/$all/$around",
            'win32': "%zoom%/%all%/%around%"
        }
        platform_tests['linux2'] = platform_tests['linux']
        platform_tests['darwin'] = platform_tests['linux']
        win_expected = os.path.join(os.getcwd(), "come\\back\\down")
        win_expected = win_expected.replace(os.path.sep, '/')
        platform_expected = {
            'linux': os.path.join(os.getcwd(), "come/back/down"),
            'win32': win_expected
        }
        platform_expected['linux2'] = platform_expected['linux']
        platform_expected['darwin'] = platform_expected['linux']
        env_vars = {
            'zoom': 'come',
            'all': 'back',
            'around': 'down'
        }
        os.environ.update(env_vars)
        result = nxt_path.full_file_expand(platform_tests[sys.platform])
        self.assertEqual(result, platform_expected[sys.platform])

    def test_cross_platform_expansion(self):
        tests = [
            '~/%FOO%/$BAR/%BAZ%',
            '~/$FOO/%BAR%/$BAZ'
        ]
        os.environ['FOO'] = 'FIRST'
        os.environ['BAR'] = 'SECOND'
        os.environ['BAZ'] = 'THIRD'
        expected = os.path.expanduser('~/FIRST/SECOND/THIRD').replace(os.sep,
                                                                      '/')
        for test_str in tests:
            result = nxt_path.full_file_expand(test_str)
            self.assertEqual(expected, result)



class TestNodePathing(unittest.TestCase):
    def test_path_partition(self):
        test_in_exp = {
            'stuff.thing': ('stuff', 'thing'),
            'all/other/stuff.another': ('all/other/stuff', 'another'),
            '/stuff/and/things': ('/stuff/and/things', None)
        }
        for inp, exp in test_in_exp.items():
            result = nxt_path.path_attr_partition(inp)
            self.assertEqual(exp, result)

    def test_namespace_build(self):
        test_in_exp = {
            'stuff.thing': ['stuff'],
            'all/other/stuff.another': ['all', 'other', 'stuff'],
            '/stuff/and/things': ['stuff', 'and', 'things']
        }
        for inp, exp in test_in_exp.items():
            result = nxt_path.str_path_to_node_namespace(inp)
            self.assertEqual(exp, result)

    def test_path_build(self):
        test_exp_in = {
            '/stuff': ['stuff'],
            '/all/other/stuff': ['all', 'other', 'stuff'],
            '/stuff/and/things': ['stuff', 'and', 'things']
        }
        for exp, inp in test_exp_in.items():
            result = nxt_path.node_namespace_to_str_path(inp)
            self.assertEqual(exp, result)

    def test_partition_helpers(self):
        test_in_exp = {
            'stuff.thing': 'stuff',
            'all/other/stuff.another': 'all/other/stuff',
            '/stuff/and/things': '/stuff/and/things'
        }
        for inp, exp in test_in_exp.items():
            result = nxt_path.node_path_from_attr_path(inp)
            self.assertEqual(exp, result)
        test_in_exp = {
            'stuff.thing': 'thing',
            'all/other/stuff.another': 'another',
            '/stuff/and/things': None
        }
        for inp, exp in test_in_exp.items():
            result = nxt_path.attr_name_from_attr_path(inp)
            self.assertEqual(exp, result)
        test_in_exp = {
            'all/other/stuff': 'stuff',
            '/stuff/and/things': 'things'
        }
        for inp, exp in test_in_exp.items():
            result = nxt_path.node_name_from_node_path(inp)
            self.assertEqual(exp, result)
        test_in_exp = {
            'all/other/stuff': 'all/other',
            '/stuff/and/things': '/stuff/and'
        }
        for inp, exp in test_in_exp.items():
            result = nxt_path.get_parent_path(inp)
            self.assertEqual(exp, result)
        test_in_exp = {
            '/all/other/stuff': '/all',
            '/stuff/and/things': '/stuff'
        }
        for inp, exp in test_in_exp.items():
            result = nxt_path.get_root_path(inp)
            self.assertEqual(exp, result)

    def test_make_attr_path(self):
        test_exp_in = {
            'stuff.thing': ('stuff', 'thing'),
            'all/other/stuff.another': ('all/other/stuff', 'another')
        }
        for exp, inp in test_exp_in.items():
            result = nxt_path.make_attr_path(inp[0], inp[1])
            self.assertEqual(exp, result)

    def test_relative_node_paths(self):
        os.chdir(os.path.dirname(__file__))
        test_stage = Session().load_file(filepath='RelPathRef.nxt')
        comp_layer = test_stage.build_stage()
        child_1_path = '/parent/child1'
        child_1 = test_stage.top_layer.lookup(child_1_path)

        test_in_exp = {
            'grandchild1': '/parent/child1/grandchild1',
            '..': '/parent',
            '../child2': '/parent/child2',
            'grandchild1/../../child2': '/parent/child2',
            './././././grandchild1/./././././.': '/parent/child1/grandchild1'
        }
        for inp, exp in test_in_exp.items():
            result = nxt_path.expand_relative_node_path(inp, child_1_path)
            self.assertEqual(exp, result)
        print("Test that a relative node path works as an instance path")
        inst_target = comp_layer.lookup('/parent/inst_target')
        inst_path = getattr(inst_target, INTERNAL_ATTRS.INSTANCE_PATH)
        inst_source_node = comp_layer.lookup(inst_path)
        self.assertIsNotNone(inst_source_node)
        start_node = '/some/node'
        not_rel = '/another/node'
        result = nxt_path.expand_relative_node_path(not_rel, start_node)
        self.assertEqual(result, not_rel)

    def test_depth_trimming(self):
        inp = '/keep/it/stupid/simple'
        exp = '/'
        result = nxt_path.trim_to_depth(inp, 0)
        self.assertEqual(exp, result)
        exp = '/keep'
        result = nxt_path.trim_to_depth(inp, 1)
        self.assertEqual(exp, result)
        exp = '/keep/it/stupid'
        result = nxt_path.trim_to_depth(inp, 3)
        self.assertEqual(exp, result)
        exp = '/keep/it/stupid/simple'
        result = nxt_path.trim_to_depth(inp, 4)
        self.assertEqual(exp, result)
        exp = '/keep/it/stupid/simple'
        result = nxt_path.trim_to_depth(inp, 7)
        self.assertEqual(exp, result)



# Built-in
from functools import partial
from collections import OrderedDict
import logging
import json
import copy
import sys

# Internal
if sys.version_info[0] == 2:
    import nxt_path
    import nxt_io
else:
    from . import nxt_path
    from . import nxt_io

logger = logging.getLogger(__name__)


class INTERNAL_ATTRS(object):
    """Class uses for easy access to internal attr constants and various
    lists of them used for composition. Only modify if you are sure of the
    ramifications to the comp engine.
    The _prefix constant should be added to every attr name to match our
    convention. When an internal attr is saved the prefix is stripped.
    """
    _prefix = '_'
    CHILD_ORDER = _prefix + 'child_order'
    COMMENT = _prefix + 'comment'
    COMPUTE = _prefix + 'code'
    ENABLED = _prefix + 'enabled'
    EXECUTE_IN = _prefix + 'execute_in'
    INSTANCE_PATH = _prefix + 'instance'
    NODE_PATH = _prefix + 'node_path'
    NAME = _prefix + 'name'
    PARENT_PATH = _prefix + 'parent_path'
    PROXY = _prefix + 'proxy'
    SOURCE_LAYER = _prefix + 'source_layer'
    START_POINT = _prefix + 'start_point'
    CACHED_CODE = _prefix + 'cached_code'
    # List of python attrs that a node will have but we don't want to parse or
    # considering in our composite logic
    BUILTINS = tuple(dir(type('NodeSpec', (object,), {})))
    # A list of node attrs that are used internally in our composite logic but
    # tracked and like user attrs with a `_source__nxt` meta attr
    TRACKED = (COMPUTE, EXECUTE_IN, COMMENT, START_POINT, ENABLED,
               INSTANCE_PATH)
    # Un-tracked internal attrs
    UNTRACKED = (CHILD_ORDER, NAME, SOURCE_LAYER, PROXY, PARENT_PATH,
                 NODE_PATH)
    # Attrs names that must be reserved but are not used in the comp engine
    NO_COMP = (CACHED_CODE,)
    # The full tuple of all internal attr names
    PROTECTED = BUILTINS + TRACKED + UNTRACKED + NO_COMP
    # Tuple for easy looping when saving/loading node data
    ALL = TRACKED + UNTRACKED
    # Defines order of node attrs in the save file
    SAVED = (START_POINT, INSTANCE_PATH, EXECUTE_IN, CHILD_ORDER, ENABLED,
             COMMENT, COMPUTE)
    WORLD_NODE_SAVED = (COMMENT, COMPUTE)
    ALLOW_NO_OPINION = (ENABLED,)
    # Dict mapping internal attr to a partial object that generates a default
    # for the given attr
    DEFAULTS = {COMPUTE: partial(list, ()), PARENT_PATH: nxt_path.WORLD}
    REQUIRES_RECOMP = tuple(set(ALL) - {CHILD_ORDER, INSTANCE_PATH, COMPUTE,
                                        ENABLED})


    @classmethod
    def as_save_key(cls, attr):
        key = attr.partition(INTERNAL_ATTRS._prefix)[-1]
        if attr == cls.NAME:
            key = 'name'
        return key


class META_ATTRS(object):
    _prefix = INTERNAL_ATTRS._prefix
    _suffix = '__nxt'
    COMMENT = _prefix + 'comment' + _suffix
    TYPE = _prefix + 'type' + _suffix
    SOURCE = _prefix + 'source' + _suffix
    VALUE = 'value'
    ALL = (COMMENT, TYPE, SOURCE)

    @classmethod
    def as_save_key(cls, key):
        if key == cls.VALUE:
            return key
        return key.partition(cls._prefix)[-1].partition(cls._suffix)[0]


class Node(object):
    _name = 'node'
    _parent_path = nxt_path.WORLD
    _child_order = []
    _proxy = False
    _enabled = True


def create_spec_node(node_data, layer, parent_path=nxt_path.WORLD,
                     is_proxy=False):
    attrs = {}
    node_name = node_data[INTERNAL_ATTRS.as_save_key(INTERNAL_ATTRS.NAME)]
    path = nxt_path.join_node_paths(parent_path, node_name)
    # Parse for attrs
    user_attrs = node_data.get(nxt_io.SAVE_KEY.ATTRS, {})
    for attr_name, data in user_attrs.items():
        attrs[attr_name] = None
        attrs[attr_name + META_ATTRS.SOURCE] = (layer.real_path, path)
        for sub_attr, value in data.items():
            if sub_attr == META_ATTRS.VALUE:
                attrs[attr_name] = value
                continue
            meta_attr = META_ATTRS._prefix + sub_attr + META_ATTRS._suffix
            if meta_attr not in META_ATTRS.ALL:
                logger.warning('Invalid meta attr "{}"'.format(meta_attr))
                logger.warning('Expected {}'.format(META_ATTRS.ALL))
                continue
            full_attr_name = (attr_name + meta_attr)
            attrs[full_attr_name] = value
    for attr in INTERNAL_ATTRS.ALL:
        key = INTERNAL_ATTRS.as_save_key(attr)
        val = node_data.get(key)
        if (not has_opinion(val) and attr not in
                INTERNAL_ATTRS.ALLOW_NO_OPINION):
            val = INTERNAL_ATTRS.DEFAULTS.get(attr)
            if callable(val):
                val = val()
        attrs[attr] = val
    # Remove duplicates in child order
    child_order = []
    for c in node_data.get('child_order', []):
        if c not in child_order:
            child_order += [c]
    attrs.update({INTERNAL_ATTRS.NAME: node_name,
                  INTERNAL_ATTRS.PROXY: is_proxy,
                  INTERNAL_ATTRS.CHILD_ORDER: child_order,
                  INTERNAL_ATTRS.SOURCE_LAYER: layer.real_path
                  })
    pp_key = INTERNAL_ATTRS.as_save_key(INTERNAL_ATTRS.PARENT_PATH)
    pp_attr = INTERNAL_ATTRS.PARENT_PATH
    if is_proxy:
        attrs[pp_attr] = node_data.get(pp_key) or parent_path
    else:
        attrs[pp_attr] = parent_path

    node_spec = SpecNode.new(attrs=attrs)
    parent_object = layer.lookup(parent_path)
    layer.clear_node_child_cache(path)
    layer.clear_node_child_cache(parent_path)
    if parent_object:
        proxies = (getattr(node_spec, INTERNAL_ATTRS.PROXY) or
                    getattr(parent_object, INTERNAL_ATTRS.PROXY))
        child_order = getattr(parent_object, INTERNAL_ATTRS.CHILD_ORDER)
        if node_name not in child_order and not proxies:
            new_co = list_merger(child_order, [node_name])
            setattr(parent_object, INTERNAL_ATTRS.CHILD_ORDER, new_co)
    for attr in INTERNAL_ATTRS.TRACKED:
        setattr(node_spec, attr + META_ATTRS.SOURCE,
                (layer.real_path, path))
    return node_spec


class SpecNode(object):
    @classmethod
    def new(cls, attrs=None):
        if attrs is None:
            attrs = {}
        return type(cls.__name__, (Node,), attrs)


class CompNode(object):

    @classmethod
    def new(cls, spec_node, attrs=None):
        bases = (spec_node,)
        if attrs is None:
            attrs = {}
        return type(cls.__name__, bases, attrs)


def get_node_attr(node_object, attr, default=None):
    return getattr(node_object, attr, default)


def get_opinion(node, attr):
    """Check if the node attr has an opinion.
    Returns a tuple: (attr_value, bool_has_opinion)
    :param node: NxtNode
    :param attr: String of attr name
    :return: (attr_value, bool_has_opinion)
    """
    attr_value = get_node_attr(node, attr)
    op_exists = has_opinion(attr_value)
    return attr_value, op_exists


def get_node_enabled(node):
    """Get the enabled state of the given node. If the enabled state is not
    set it (None) the default return is True.
    :param node: Node object
    :return: Bool of enabled state
    """
    return getattr(node, INTERNAL_ATTRS.ENABLED, None)


def has_opinion(attr_value):
    return attr_value not in (None, [])


def has_stronger_opinion(comp_node, attr, target_layer):
    """Check if the given attr on the given node has a stronger opinion than
    the opinion on the target layer (if any). We check the source layers of
    each base class. We are looping strong to weak so if there was no
    opinion before hitting the target layer there isn't a stronger opinion.
    :param comp_node: NxtCompNode
    :param attr: String of attr name
    :param target_layer: NxtSpecLayer
    :return: bool (True if comp node has opinion that would overload the
    target layer's opinion)
    """
    result = False
    comp_opinion, _ = get_opinion(comp_node, attr)
    val = None
    for b in comp_node.__bases__:
        source_layer = getattr(b, INTERNAL_ATTRS.SOURCE_LAYER)
        val, has = get_opinion(b, attr)
        if source_layer == target_layer:
            break
        result = has
        if has:
            break
    if comp_opinion == val:
        return False
    return result


def get_node_as_dict(spec_node):
    """Get a spec node as an ordered dictionary. The return data is
    strcutred as it would be in a saved graph file.
    :param spec_node: SpecNode object
    :return: OrderedDict
    """
    user_attrs = OrderedDict()
    spec_dict = OrderedDict()
    # user attrs
    for key in spec_node.__dict__.keys():
        if key in INTERNAL_ATTRS.PROTECTED or key.endswith(META_ATTRS.ALL):
            continue
        # Sub-attrs
        comment = get_node_attr(spec_node, key + META_ATTRS.COMMENT)
        typ = get_node_attr(spec_node, key + META_ATTRS.TYPE)
        value = get_node_attr(spec_node, key)
        sub_attrs = OrderedDict()
        if comment:
            sub_attrs[META_ATTRS.as_save_key(META_ATTRS.COMMENT)] = comment
        if typ:
            sub_attrs[META_ATTRS.as_save_key(META_ATTRS.TYPE)] = typ
        if value:
            try:  # Marshal data
                json.dumps(value)
                val_copy = copy.deepcopy(value)
                sub_attrs[META_ATTRS.as_save_key(META_ATTRS.VALUE)] = val_copy
            except TypeError:  # value is not JSON serializable
                pass
        sorted_sub_attrs = OrderedDict(sorted(sub_attrs.items(),
                                              key=lambda x: x[0]))
        user_attrs[key] = sorted_sub_attrs
    if user_attrs:
        sorted_attrs = OrderedDict(sorted(user_attrs.items(),
                                          key=lambda x: x[0]))
        spec_dict[nxt_io.SAVE_KEY.ATTRS] = sorted_attrs
    # Saving of internal attrs
    save_attrs = INTERNAL_ATTRS.SAVED
    if getattr(spec_node, INTERNAL_ATTRS.NAME, None) == nxt_path.WORLD:
        save_attrs = INTERNAL_ATTRS.WORLD_NODE_SAVED
    for attr in save_attrs:
        value, has_opinion = get_opinion(spec_node, attr)
        if not has_opinion:
            continue
        save_attr_name = INTERNAL_ATTRS.as_save_key(attr)
        spec_dict[save_attr_name] = copy.deepcopy(value)
    return _order_node_dict(spec_dict)


def _order_node_dict(node_dict):
    """Given a node in dictionary format, order the node dictionary based on
    INTERNAL_ATTRS.SAVED order. Any keys in given dictionary not found will
    be at the end.
    :param node_dict: node in dict format
    :type node_dict: dict
    """
    result = OrderedDict()
    user_attrs = node_dict.get(nxt_io.SAVE_KEY.ATTRS)
    remaining_keys = list(node_dict.keys())
    for internal_attr in INTERNAL_ATTRS.SAVED:
        if internal_attr == INTERNAL_ATTRS.COMPUTE and user_attrs:
            result[nxt_io.SAVE_KEY.ATTRS] = user_attrs
        save_key = INTERNAL_ATTRS.as_save_key(internal_attr)
        if save_key not in node_dict:
            continue
        remaining_keys.remove(save_key)
        result[save_key] = node_dict[save_key]
    for key in remaining_keys:
        result[key] = node_dict[key]
    return result


def get_node_path(node):
    """Only to be used for inferred nodes who do not belong to a layer.
    :param node: Node object
    :return: string of node path
    """
    return nxt_path.join_node_paths(getattr(node, INTERNAL_ATTRS.PARENT_PATH),
                                    getattr(node, INTERNAL_ATTRS.NAME))


def list_merger(base_list, overlay_list):
    """It is important to note we expect both arg lists to be non-repeating.
    Merges lists such that the `base_list` is preserved and
    the `overlay_list` is overlaid onto it.
    Example:
    source = ['Node', 'Node1']
    target = ['Node', 'Jack']
    returns = ['Node', 'Jack', 'Node1']
    :param base_list: Source list of object we want to overlay.
    :param overlay_list: Target list of object we want overlay onto.
    :return: Non-repeating list of object.
    """
    merged_list = base_list[:]
    result_len = len(merged_list)
    if base_list == overlay_list:
        return merged_list
    src_set = set(base_list)
    tgt_set = set(overlay_list)
    tgt_order = overlay_list[:]
    src_order = base_list[:]
    if not tgt_set.intersection(src_set):
        only_source = [n for n in src_order if n not in tgt_order]
    else:
        only_source = []

    if tgt_set.issubset(src_set):
        return merged_list
    idx = 0
    for item in tgt_order:
        if item not in src_order:
            offset = 0
            try:
                _ = src_order[idx]
                valid = True
            except IndexError:
                valid = False
            if valid:
                prev_idx = max(0, idx - 1)
                match = src_order[prev_idx] == tgt_order[prev_idx]
                if prev_idx >= 0 and match:
                    offset = 0
                elif match:
                    offset = prev_idx
                else:
                    next_idx = idx + 1
                    try:
                        match = src_order[next_idx] == tgt_order[next_idx]
                        if match:
                            offset = next_idx
                    except IndexError:
                        pass
                insert = idx + offset
                if insert >= 0 or result_len < 3:
                    merged_list.insert(insert, item)
                else:
                    merged_list.append(item)
            else:
                merged_list.insert(idx, item)
            result_len += 1
        idx += 1
    for item in reversed(only_source):
        merged_list.remove(item)
        merged_list.insert(0, item)
    return merged_list



"""Methods for file input and output of nxt file data.
"""
# Builtin
import json
import logging
import os
import time
import sys
# Python 2/3 compatibility
if sys.version_info[0] == 2:
    import cPickle as pickle
    # Internal
    import nxt_path
    import clean_json
    import legacy
    from constants import FILE_FORMAT, GRAPH_VERSION
    from nxt_layer import SAVE_KEY
else:
    import pickle
    # Internal
    from . import nxt_path
    from . import clean_json
    from . import legacy
    from .constants import FILE_FORMAT, GRAPH_VERSION
    from .nxt_layer import SAVE_KEY

import gc
import tempfile

nxt_folder = os.path.dirname(os.path.abspath(__file__))
BUILTIN_GRAPHS_DIR = os.path.join(nxt_folder, 'builtin')
BUILTIN_GRAPHS_ENV_VAR = 'NXT_BUILTINS'
os.environ[BUILTIN_GRAPHS_ENV_VAR] = BUILTIN_GRAPHS_DIR
logger = logging.getLogger(__name__)

plugin_expanders = []


def register_reference_path_expander(expander):
    """Register a method to expand reference paths. An expander is a callable \
    that takes exactly 1 argument, a filepath.

    :param expander: A python callable that takes a single argument, and \
    returns an expanded filepath.
    :type expander: callable
    """
    global plugin_expanders
    msg = 'Registered reference path expander from ' + str(expander.__module__)
    logger.info(msg)
    plugin_expanders += [expander]


def load_file_data(filepath):
    """Given a file path this function determines if its a known nxt save
    format and attempts to open it. If the file is out of date it is passed
    to the legacy converter for conversion.

    :param filepath: string of save file filepath
    :type filepath: str
    :return: dict of file data
    """
    real_path = nxt_path.full_file_expand(filepath)
    for expander in plugin_expanders:
        found_path = expander(filepath)
        if not os.path.isfile(found_path):
            continue
        real_path = found_path
        break
    _, ext = os.path.splitext(real_path)
    if ext not in FILE_FORMAT._ALL:
        raise IOError('Unknown filetype "{}"'.format(ext))
    file_type = FILE_FORMAT.ASCII
    if file_type == FILE_FORMAT.BINARY:
        with open(real_path, 'rb') as file_object:
                file_data = pickle.load(file_object)
    else:
        with open(real_path, 'r') as file_object:
            file_data = json.load(file_object, object_hook=clean_json._byteify)
    file_version = file_data.get(SAVE_KEY.VERSION)
    if not file_version:
        file_version = GRAPH_VERSION.VERSION_STR
        logger.warning('Assuming file version `{}` if trying to open a '
                       'file from before 1.0.0 please add '
                       'the following line to to top of your save file: '
                       '"version": "0.45.0"'.format(GRAPH_VERSION.VERSION_STR))
    if not file_version.count('.'):
        raise IOError('Invalid version number format! Please ensure your save '
                      'file has the following line: "version": "1.x" where '
                      'the x is replaced with an int.')
    # Strip off patch number as we don't support save converters for hot fixes
    major_version, minor_version = [int(n) for n in
                                    file_version.split('.')[0:2]]
    file_too_new = (major_version > GRAPH_VERSION.MAJOR or
                    (major_version == GRAPH_VERSION.MAJOR
                     and minor_version > GRAPH_VERSION.MINOR))
    if file_too_new:
        raise IOError('You are attempting to open a file saved with a newer '
                      'version of nxt ({}). '
                      'Please update nxt!'.format(file_version))
    if major_version == 0 or minor_version < GRAPH_VERSION.MINOR:
        # Calls a legacy check if the full version number doesn't match the
        # current version
        start = time.time()
        template = (file_data.get(SAVE_KEY.VERSION),
                    GRAPH_VERSION.VERSION_STR, str(round(time.time() - start, 3)))
        file_data = legacy.FileConverter.get_converted_data(file_data)
        logger.info("File converted from '{}' to '{}' in {}".format(*template))
    file_data[SAVE_KEY.FILEPATH] = filepath
    file_data[SAVE_KEY.REAL_PATH] = real_path
    name_from_file = os.path.splitext(os.path.basename(filepath))[0]
    file_data[SAVE_KEY.NAME] = file_data.get(SAVE_KEY.NAME, name_from_file)
    return file_data


def save_file_data(save_data, filepath=None, file_format=FILE_FORMAT.ASCII):
    """Saves the given file data to the given file path in the given file
    format. If no filepath is provided a temp file is generated.
    The default file format is ASCII.

    :param save_data: dict of save data
    :param filepath: string of desired output file path or None
    :param file_format: FILE_FORMAT constant
    :return: string of output file path.
    """
    start = time.time()
    ext = file_format
    _, filepath_ext = os.path.splitext(filepath)
    if filepath_ext not in FILE_FORMAT._ALL:
        raise TypeError('Unknown file extension "{}"'.format(filepath_ext))
    elif filepath_ext != ext:
        ext = filepath_ext
    filepath = filepath or tempfile.mkstemp(prefix='nxt_tmp_',
                                            suffix=ext)[1]
    filepath = filepath.replace(os.sep, '/')
    if not filepath.endswith(ext):
        filepath += ext
    if file_format == FILE_FORMAT.BINARY:
        gc.disable()
        with open(filepath, 'wb') as out_file:
            pickle.dump(save_data, out_file, protocol=-1)
        gc.enable()
    else:
        with open(filepath, 'w') as out_file:
            json.dump(save_data, out_file, indent=4, sort_keys=False,
                      separators=(',', ': '))
    logger.info('Successfully saved "' + filepath + '"')
    update_time = str(int(round((time.time() - start) * 1000)))
    logger.debug("Saved in: " + update_time + "ms")
    return filepath


def generate_temp_file(suffix='.nxt'):
    """Safely generates a temp file and returns a Windows safe path

    :param suffix: Optional suffix for the file, default is `.nxt`
    :return: String filepath
    """
    _, cache_filepath = tempfile.mkstemp(suffix=suffix)
    return cache_filepath.replace(os.sep, '/')


def generate_temp_dir(prefix='nxt_tmp_'):
    """Safely generates a temp dir and returns a Windows safe path

    :param prefix: Optional prefix to the dir name, default is `nxt_tmp_`
    :return: String dir path
    """
    return tempfile.mkdtemp(prefix=prefix)



from nxt.tokens import register_token

PREFIX = 'ex::'


def detect_token_type(value):
    return value.startswith(PREFIX)


def resolve_token(stage, node, value, layer, **kwargs):
    value = stage.resolve(node, value, layer, **kwargs)
    # Reverses given value
    return value[::-1]


register_token(PREFIX, detect_token_type, resolve_token)



import sys


def load(json_load):
    return _byteify(json_load)


def _byteify(data, ignore_dicts=False):
    '''
    Encodes any unicode data as utf-8
    :param data: A data object from json.load or json.loads
    :param ignore_dicts: Should be true if loading json without a top level dict
    :return: unicode/byte free data
    '''
    # Python 2/3 compatibility
    if sys.version_info[0] == 2:
        if isinstance(data, unicode):
            return data.encode('utf-8')
    else:
        if isinstance(data, bytes):
            return data.encode('utf-8')
    if isinstance(data, list):
        return [_byteify(item, ignore_dicts=True) for item in data]
    if isinstance(data, dict) and not ignore_dicts:
        return {_byteify(key, ignore_dicts=True): _byteify(value, ignore_dicts=True) for key, value in data.items()}
    return data
